# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import mock

from oslo_config import cfg
from oslo_utils import uuidutils
from sqlalchemy.orm import exc
import unittest
import webob.exc

from neutron.db import l3_db
from neutron.tests import fake_notifier

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import cb_events as events
from networking_cisco.backwards_compatibility import cb_registry as registry
from networking_cisco.backwards_compatibility import cb_resources as resources
from networking_cisco.backwards_compatibility import l3_const
import networking_cisco.plugins
from networking_cisco.plugins.cisco.common import (
    cisco_constants as cisco_const)
from networking_cisco.plugins.cisco.db.l3 import ha_db
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.tests.unit.cisco.device_manager import (
    device_manager_test_support)
from networking_cisco.tests.unit.cisco.l3 import (
    test_l3_router_appliance_plugin)


_uuid = uuidutils.generate_uuid

EXTERNAL_GW_INFO = l3_const.EXTERNAL_GW_INFO

CORE_PLUGIN_KLASS = device_manager_test_support.CORE_PLUGIN_KLASS
DEFAULT_PRIORITY = ha_db.DEFAULT_MASTER_PRIORITY
L3_PLUGIN_KLASS = (
    "networking_cisco.tests.unit.cisco.l3.test_ha_l3_router_appliance_plugin."
    "TestApplianceHAL3RouterServicePlugin")
extensions_path = networking_cisco.plugins.__path__[0] + '/cisco/extensions'

DEVICE_OWNER_ROUTER_INTF = bc.constants.DEVICE_OWNER_ROUTER_INTF


def _sort_routes(routes):
    return sorted(routes, key=lambda route: route['destination'])


class TestHAL3RouterApplianceExtensionManager(
        test_l3_router_appliance_plugin.TestL3RouterApplianceExtensionManager):

    def get_resources(self):
        l3_const.RESOURCE_ATTRIBUTE_MAP['routers'].update(
            ha.EXTENDED_ATTRIBUTES_2_0['routers'])
        return super(TestHAL3RouterApplianceExtensionManager,
                     self).get_resources()


# A set routes and HA capable L3 routing service plugin class
# supporting appliances
class TestApplianceHAL3RouterServicePlugin(
    ha_db.HA_db_mixin,
        test_l3_router_appliance_plugin.TestApplianceL3RouterServicePlugin):

    supported_extension_aliases = (
        test_l3_router_appliance_plugin.TestApplianceL3RouterServicePlugin.
        supported_extension_aliases + [ha.HA_ALIAS])

    def cleanup_after_test(self):
        """Reset all class variables to their default values.
        This is needed to avoid tests to pollute subsequent tests.
        """
        TestApplianceHAL3RouterServicePlugin._router_schedulers = {}
        TestApplianceHAL3RouterServicePlugin._router_drivers = {}
        TestApplianceHAL3RouterServicePlugin._namespace_router_type_id = None
        TestApplianceHAL3RouterServicePlugin._backlogged_routers = set()
        TestApplianceHAL3RouterServicePlugin._refresh_router_backlog = True


# TODO(bobmel): Add tests that ensures that Cisco HA is not applied on
# Namespace-based routers
class HAL3RouterApplianceNamespaceTestCase(
        test_l3_router_appliance_plugin.L3RouterApplianceNamespaceTestCase):

    _is_ha_tests = True

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        if ext_mgr is None:
            ext_mgr = TestHAL3RouterApplianceExtensionManager()
        super(HAL3RouterApplianceNamespaceTestCase, self).setUp(
            l3_plugin=l3_plugin, ext_mgr=ext_mgr)


class HAL3RouterTestsMixin(object):

    def _get_ha_defaults(self, ha_enabled=None, ha_type=None,
                         redundancy_level=None, priority=DEFAULT_PRIORITY,
                         state=ha.HA_ACTIVE, probing_enabled=None,
                         probe_target=None, probe_interval=None):

        if ha_enabled is None:
            ha_enabled = self._is_ha_tests
        if not ha_enabled:
            return {ha.ENABLED: False}
        ha_details = {
            ha.TYPE: ha_type or cfg.CONF.ha.default_ha_mechanism,
            ha.PRIORITY: priority,
            ha.STATE: state,
            ha.REDUNDANCY_LEVEL: (redundancy_level or
                                  cfg.CONF.ha.default_ha_redundancy_level),
            ha.PROBE_CONNECTIVITY: (
                probing_enabled if probing_enabled is not None else
                cfg.CONF.ha.connectivity_probing_enabled_by_default)}
        if probing_enabled:
            ha_details.update({
                ha.PROBE_TARGET: (probe_target or
                                  cfg.CONF.ha.default_ping_target),
                ha.PROBE_INTERVAL: (probe_interval or
                                    cfg.CONF.ha.default_ping_interval)})
        return {ha.ENABLED: ha_enabled, ha.DETAILS: ha_details}

    def _verify_ha_settings(self, router, expected_ha):
            self.assertEqual(router[ha.ENABLED], expected_ha[ha.ENABLED])
            if expected_ha[ha.ENABLED]:
                if ha.DETAILS in expected_ha:
                    ha_details = copy.deepcopy(router[ha.DETAILS])
                    redundancy_routers = ha_details.pop(ha.REDUNDANCY_ROUTERS)
                    self.assertDictEqual(ha_details,
                                         expected_ha[ha.DETAILS])
                    self.assertEqual(
                        len(redundancy_routers),
                        expected_ha[ha.DETAILS][ha.REDUNDANCY_LEVEL])
                else:
                    self.assertTrue(ha.DETAILS not in router)
            else:
                self.assertIsNone(router.get(ha.DETAILS))

    def _verify_router_gw_port(self, router_id, external_net_id,
                               external_subnet_id):
        body = self._list('ports',
                          query_params='device_id=%s' % router_id)
        ports = body['ports']
        self.assertEqual(len(ports), 1)
        p_e = ports[0]
        self.assertEqual(p_e['network_id'], external_net_id)
        self.assertEqual(p_e['fixed_ips'][0]['subnet_id'], external_subnet_id)
        self.assertEqual(p_e['device_owner'],
                         bc.constants.DEVICE_OWNER_ROUTER_GW)


class HAL3RouterApplianceVMTestCase(
    HAL3RouterTestsMixin,
        test_l3_router_appliance_plugin.L3RouterApplianceVMTestCase):

    _is_ha_tests = True

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        if ext_mgr is None:
            ext_mgr = TestHAL3RouterApplianceExtensionManager()
        cfg.CONF.set_override('default_ha_redundancy_level', 2, group='ha')
        super(HAL3RouterApplianceVMTestCase, self).setUp(
            l3_plugin=l3_plugin, ext_mgr=ext_mgr)

    def test_router_add_interface_port(self):
        orig_update_port = self.core_plugin.update_port
        with self.router() as router, (
            self.port()) as port, (
                mock.patch.object(self.core_plugin,
                                  'update_port')) as update_port:
            update_port.side_effect = orig_update_port
            r = router['router']
            p = port['port']
            body = self._router_interface_action('add', r['id'], None, p['id'])
            self.assertIn('port_id', body)
            self.assertEqual(p['id'], body['port_id'])
            r_ids = [rr['id']
                     for rr in [r] + r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]]
            # get ports for the user visible router and its two redundancy
            # routers for which device_id attribute should have been updated
            params = "&".join(["device_id=%s" % r_id for r_id in r_ids])
            router_ports = self._list('ports', query_params=params)['ports']
            self.assertEqual(len(router_ports), 3)
            expected_calls = []
            for port in router_ports:
                expected_port_update = {
                    'port': {'device_owner': DEVICE_OWNER_ROUTER_INTF,
                             'device_id': port['device_id']}}
                expected_calls.append(
                    mock.call(mock.ANY, port['id'], expected_port_update))
            update_port.assert_has_calls(expected_calls, any_order=True)
            r_port_ids = {r_p['id'] for r_p in router_ports}
            # get the extra port for the user visible router
            other_router_ports = [r_p for r_p in self._list('ports')['ports']
                                  if r_p['id'] not in r_port_ids]
            # clean-up
            self._router_interface_action('remove', r['id'], None, p['id'])

            # should only be one since we've created one router port
            self.assertEqual(1, len(other_router_ports))
            self.assertEqual(DEVICE_OWNER_ROUTER_INTF,
                             other_router_ports[0]['device_owner'])

    def test_hidden_port_creation_includes_dns_attribute(self):
        with mock.patch('networking_cisco.plugins.cisco.db.l3.ha_db.'
                        'utils.is_extension_supported',
                        return_value=True) as extension_support,\
                mock.patch.object(self.core_plugin, 'create_port') as c_p_mock:
                    # Verify the function is called with extension dns enabled
                    expected_calls = [
                        mock.call(self.core_plugin, 'dns-integration')]
                    self.l3_plugin._create_hidden_port(
                        'fake_ctx', 'some_network_id', 'some_device_id', [])
                    extension_support.assert_has_calls(expected_calls,
                                                       any_order=True)
                    c_p_mock.assert_called_once_with('fake_ctx', mock.ANY)
                    self.assertEqual(
                        '', c_p_mock.call_args[0][1]['port']['dns_name'])

    def _test_create_ha_router(self, router, subnet, ha_settings=None):
        if ha_settings is None:
            ha_settings = self._get_ha_defaults()
        self._verify_ha_settings(router, ha_settings)
        if subnet is not None:
            self.assertEqual(subnet['network_id'],
                             router['external_gateway_info']['network_id'])
            self._verify_router_gw_port(router['id'], subnet['network_id'],
                                        subnet['id'])
        ha_disabled_settings = self._get_ha_defaults(ha_enabled=False)
        # verify redundancy routers
        for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
            rr = self._show('routers', rr_info['id'])
            # check that redundancy router is hidden
            self.assertEqual(rr['router']['tenant_id'], '')
            # redundancy router should have ha disabled
            self._verify_ha_settings(rr['router'], ha_disabled_settings)
            if subnet is not None:
                # check that redundancy router has all ports
                self._verify_router_gw_port(rr['router']['id'],
                                            subnet['network_id'], subnet['id'])

    def test_create_ha_router_with_defaults(self):
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s['subnet']['network_id']}) as r:
                self._test_create_ha_router(r['router'], s['subnet'])

    def test_create_non_gw_ha_router_with_defaults(self):
        with self.router() as r:
            self._test_create_ha_router(r['router'], None)

    def test_create_ha_router_with_defaults_non_admin_succeeds(self):
        tenant_id = _uuid()
        with self.network(tenant_id=tenant_id) as n_external:
            res = self._create_subnet(self.fmt, n_external['network']['id'],
                                      cidr='10.0.1.0/24', tenant_id=tenant_id)
            s = self.deserialize(self.fmt, res)
            self._set_net_external(s['subnet']['network_id'])
            with self.router(
                    tenant_id=tenant_id,
                    external_gateway_info={
                        'network_id': s['subnet']['network_id']},
                    set_context=True) as r:
                self.assertEqual(
                    s['subnet']['network_id'],
                    r['router']['external_gateway_info']['network_id'])
                self.assertTrue(r['router'][ha.ENABLED])
                # non-admin users should not see ha detail
                self.assertIsNone(r['router'].get(ha.DETAILS))

    def test_create_non_gw_ha_router_with_defaults_non_admin_succeeds(self):
        tenant_id = _uuid()
        with self.router(tenant_id=tenant_id, set_context=True) as r:
            self.assertTrue(r['router'][ha.ENABLED])
            # non-admin users should not see ha detail
            self.assertIsNone(r['router'].get(ha.DETAILS))

    def test_create_ha_router_with_ha_specification(self):
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            ha_settings = self._get_ha_defaults(
                ha_type=ha.HA_GLBP, priority=DEFAULT_PRIORITY,
                probing_enabled=True,
                probe_interval=3, probe_target='10.5.5.2')
            kwargs = {ha.DETAILS: ha_settings[ha.DETAILS],
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            with self.router(arg_list=(ha.DETAILS,), **kwargs) as r:
                self._test_create_ha_router(r['router'], s['subnet'],
                                            ha_settings)

    def test_create_non_gw_ha_router_with_ha_specification(self):
        ha_settings = self._get_ha_defaults(
            ha_type=ha.HA_GLBP, priority=DEFAULT_PRIORITY,
            probing_enabled=True,
            probe_interval=3, probe_target='10.5.5.2')
        kwargs = {ha.DETAILS: ha_settings[ha.DETAILS]}
        with self.router(arg_list=(ha.DETAILS,), **kwargs) as r:
            self._test_create_ha_router(r['router'], None, ha_settings)

    def test_create_ha_router_with_ha_specification_validation_fails(self):
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            ha_settings = self._get_ha_defaults(redundancy_level=5,
                ha_type=ha.HA_GLBP, priority=15, probing_enabled=True,
                probe_interval=3, probe_target='10.5.5.2')
            kwargs = {ha.ENABLED: True,
                      ha.DETAILS: ha_settings[ha.DETAILS],
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            res = self._create_router(self.fmt, _uuid(), 'ha_router1',
                                      arg_list=(ha.ENABLED,
                                                ha.DETAILS,
                                                l3_const.EXTERNAL_GW_INFO),
                                      **kwargs)
            self.assertEqual(res.status_int, webob.exc.HTTPBadRequest.code)

    def test_create_non_gw_ha_router_with_ha_specification_validation_fails(
            self):
        ha_settings = self._get_ha_defaults(redundancy_level=5,
            ha_type=ha.HA_GLBP, priority=15, probing_enabled=True,
            probe_interval=3, probe_target='10.5.5.2')
        kwargs = {ha.ENABLED: True, ha.DETAILS: ha_settings[ha.DETAILS]}
        res = self._create_router(self.fmt, _uuid(), 'ha_router1',
                                  arg_list=(ha.ENABLED, ha.DETAILS), **kwargs)
        self.assertEqual(res.status_int, webob.exc.HTTPBadRequest.code)

    def test_create_ha_router_with_ha_specification_invalid_HA_type_fails(
            self):
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            ha_settings = self._get_ha_defaults(redundancy_level=3,
                ha_type="UNKNOWN", priority=15, probing_enabled=True,
                probe_interval=3, probe_target='10.5.5.2')
            kwargs = {ha.ENABLED: True,
                      ha.DETAILS: ha_settings[ha.DETAILS],
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            res = self._create_router(self.fmt, _uuid(), 'ha_router1',
                                      arg_list=(ha.ENABLED,
                                                ha.DETAILS,
                                                l3_const.EXTERNAL_GW_INFO),
                                      **kwargs)
            self.assertEqual(res.status_int, webob.exc.HTTPBadRequest.code)

    def test_create_non_gw_ha_router_with_ha_spec_invalid_HA_type_fails(self):
        ha_settings = self._get_ha_defaults(
            redundancy_level=3, ha_type="UNKNOWN", priority=15,
            probing_enabled=True, probe_interval=3, probe_target='10.5.5.2')
        kwargs = {ha.ENABLED: True, ha.DETAILS: ha_settings[ha.DETAILS]}
        res = self._create_router(self.fmt, _uuid(), 'ha_router1',
                                  arg_list=(ha.ENABLED, ha.DETAILS), **kwargs)
        self.assertEqual(res.status_int, webob.exc.HTTPBadRequest.code)

    def test_create_ha_router_with_ha_specification_non_admin_fails(self):
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {
                ha.ENABLED: True,
                ha.DETAILS: {ha.TYPE: ha.HA_VRRP},
                l3_const.EXTERNAL_GW_INFO: {'network_id': s['subnet'][
                    'network_id']}}
            res = self._create_router(
                self.fmt, _uuid(), 'ha_router1', set_context=True,
                arg_list=(ha.ENABLED, ha.DETAILS,
                          l3_const.EXTERNAL_GW_INFO),
                **kwargs)
            self.assertEqual(res.status_int, webob.exc.HTTPForbidden.code)

    def test_create_non_gw_ha_router_with_ha_spec_non_admin_fails(self):
        kwargs = {ha.ENABLED: True, ha.DETAILS: {ha.TYPE: ha.HA_VRRP}}
        res = self._create_router(
            self.fmt, _uuid(), 'ha_router1', set_context=True,
            arg_list=(ha.ENABLED, ha.DETAILS), **kwargs)
        self.assertEqual(res.status_int, webob.exc.HTTPForbidden.code)

    def test_create_ha_router_with_disabled_ha_type_fails(self):
        cfg.CONF.set_override('disabled_ha_mechanisms', [ha.HA_VRRP],
                              group='ha')
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {
                ha.ENABLED: True,
                ha.DETAILS: {ha.TYPE: ha.HA_VRRP},
                l3_const.EXTERNAL_GW_INFO:
                    {'network_id': s['subnet']['network_id']}}
            res = self._create_router(
                self.fmt, _uuid(), 'ha_router1',
                arg_list=(ha.ENABLED, ha.DETAILS,
                          l3_const.EXTERNAL_GW_INFO),
                **kwargs)
            self.assertEqual(res.status_int, webob.exc.HTTPConflict.code)

    def test_create_non_gw_ha_router_with_disabled_ha_type_fails(self):
        cfg.CONF.set_override('disabled_ha_mechanisms', [ha.HA_VRRP],
                              group='ha')
        kwargs = {ha.ENABLED: True, ha.DETAILS: {ha.TYPE: ha.HA_VRRP}}
        res = self._create_router(
            self.fmt, _uuid(), 'ha_router1', set_context=True,
            arg_list=(ha.ENABLED, ha.DETAILS), **kwargs)
        self.assertEqual(res.status_int, webob.exc.HTTPForbidden.code)

    def test_create_ha_router_when_ha_support_disabled_fails(self):
        cfg.CONF.set_override('ha_support_enabled', False, group='ha')
        with self.subnet() as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {
                ha.ENABLED: True,
                l3_const.EXTERNAL_GW_INFO:
                    {'network_id': s['subnet']['network_id']}}
            res = self._create_router(
                self.fmt, _uuid(), 'ha_router1',
                arg_list=(ha.ENABLED,), **kwargs)
            self.assertEqual(res.status_int, webob.exc.HTTPConflict.code)

    def test_create_non_gw_ha_router_when_ha_support_disabled_fails(self):
        cfg.CONF.set_override('ha_support_enabled', False, group='ha')
        kwargs = {ha.ENABLED: True}
        res = self._create_router(
            self.fmt, _uuid(), 'ha_router1', arg_list=(ha.ENABLED,),
            **kwargs)
        self.assertEqual(res.status_int, webob.exc.HTTPConflict.code)

    def test_show_ha_router_non_admin(self):
        tenant_id = _uuid()
        with self.network(tenant_id=tenant_id) as n_external:
            res = self._create_subnet(self.fmt, n_external['network']['id'],
                                      cidr='10.0.1.0/24', tenant_id=tenant_id)
            s = self.deserialize(self.fmt, res)
            self._set_net_external(s['subnet']['network_id'])
            with self.router(tenant_id=tenant_id,
                             external_gateway_info={
                                 'network_id': s['subnet']['network_id']},
                             set_context=True) as r:
                self.assertEqual(
                    s['subnet']['network_id'],
                    r['router']['external_gateway_info']['network_id'])
                self.assertTrue(r['router'][ha.ENABLED])
                # ensure that no ha details are included
                self.assertNotIn(ha.DETAILS, r['router'])
                r_s = self._show('routers', r['router']['id'],
                                 neutron_context=bc.context.Context('',
                                                                 tenant_id))
                self.assertTrue(r_s['router'][ha.ENABLED])
                # ensure that no ha details are included
                self.assertNotIn(ha.DETAILS, r_s['router'])

    def test_show_non_gw_ha_router_non_admin(self):
        tenant_id = _uuid()
        with self.router(tenant_id=tenant_id, set_context=True) as r:
            self.assertIsNone(r['router']['external_gateway_info'])
            self.assertTrue(r['router'][ha.ENABLED])
            # ensure that no ha details are included
            self.assertNotIn(ha.DETAILS, r['router'])
            r_s = self._show('routers', r['router']['id'],
                             neutron_context=bc.context.Context('', tenant_id))
            self.assertTrue(r_s['router'][ha.ENABLED])
            # ensure that no ha details are included
            self.assertNotIn(ha.DETAILS, r_s['router'])

    def _verify_router_ports(self, router_id, external_net_id=None,
                             external_subnet_id=None, internal_net_id=None,
                             internal_subnet_id=None):
        body = self._list('ports',
                          query_params='device_id=%s' % router_id)
        ports = body['ports']
        if external_net_id:
            p_e = ports[0]
            num_ports = 1
        else:
            num_ports = 0
        if internal_net_id:
            p_i = ports[0]
            num_ports += 1
        self.assertEqual(len(ports), num_ports)
        if num_ports == 2:
            if ports[0]['network_id'] == external_net_id:
                p_e = ports[0]
                p_i = ports[1]
            else:
                p_e = ports[1]
                p_i = ports[0]
        if external_net_id:
            self.assertEqual(p_e['fixed_ips'][0]['subnet_id'],
                             external_subnet_id)
            self.assertEqual(p_e['device_owner'],
                             bc.constants.DEVICE_OWNER_ROUTER_GW)
        if internal_net_id:
            self.assertEqual(p_i['network_id'], internal_net_id)
            self.assertEqual(p_i['fixed_ips'][0]['subnet_id'],
                             internal_subnet_id)
            self.assertEqual(p_i['device_owner'],
                             bc.constants.DEVICE_OWNER_ROUTER_INTF)

    def _ha_router_port_test(self, subnet, router, port, ha_spec=None,
                             additional_tests_function=None):
        body = self._router_interface_action('add', router['id'], None,
                                             port['id'])
        self.assertIn('port_id', body)
        self.assertEqual(body['port_id'], port['id'])
        if ha_spec is None:
            ha_spec = self._get_ha_defaults()
        # verify router visible to user
        self._verify_ha_settings(router, ha_spec)
        self._verify_router_ports(router['id'], subnet['network_id'],
                                  subnet['id'], port['network_id'],
                                  port['fixed_ips'][0]['subnet_id'])
        ha_disabled_settings = self._get_ha_defaults(ha_enabled=False)
        redundancy_routers = []
        # verify redundancy routers
        for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
            rr = self._show('routers', rr_info['id'])
            redundancy_routers.append(rr['router'])
            # check that redundancy router is hidden
            self.assertEqual(rr['router']['tenant_id'], '')
            # redundancy router should have ha disabled
            self._verify_ha_settings(rr['router'], ha_disabled_settings)
            # check that redundancy router has all ports
            self._verify_router_ports(rr['router']['id'], subnet['network_id'],
                                      subnet['id'], port['network_id'],
                                      port['fixed_ips'][0]['subnet_id'])
        if additional_tests_function is not None:
            additional_tests_function(redundancy_routers)
        # clean-up
        self._router_interface_action('remove', router['id'], None, port['id'])

    def test_ha_router_add_and_remove_interface_port(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s['subnet']['network_id']}) as r:
                with self.port() as p:
                    self._ha_router_port_test(s['subnet'], r['router'],
                                              p['port'])

    def test_non_gw_ha_router_add_and_remove_interface_port(self):
        with self.router() as r, self.port() as p:
            no_subnet = {'id': None, 'network_id': None}
            self._ha_router_port_test(no_subnet, r['router'], p['port'])

    def test_ha_router_disable_ha_succeeds(self):
        def _disable_ha_tests(redundancy_routers):
            body = {'router': {ha.ENABLED: False}}
            updated_router = self._update('routers', r['router']['id'], body)
            self._verify_ha_settings(updated_router['router'],
                                     self._get_ha_defaults(ha_enabled=False))
            # verify that the redundancy routers are indeed gone
            params = "&".join(["id=%s" % rr['id'] for rr in
                               redundancy_routers])
            redundancy_routers = self._list('routers', query_params=params)
            self.assertEqual(len(redundancy_routers['routers']), 0)

        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s['subnet']['network_id']}) as r:
                with self.port() as p:
                    self._ha_router_port_test(s['subnet'], r['router'],
                                              p['port'], None,
                                              _disable_ha_tests)

    def test_non_gw_ha_router_disable_ha_succeeds(self):
        def _disable_ha_tests(redundancy_routers):
            body = {'router': {ha.ENABLED: False}}
            updated_router = self._update('routers', r['router']['id'], body)
            self._verify_ha_settings(updated_router['router'],
                                     self._get_ha_defaults(ha_enabled=False))
            # verify that the redundancy routers are indeed gone
            params = "&".join(["id=%s" % rr['id'] for rr in
                               redundancy_routers])
            redundancy_routers = self._list('routers', query_params=params)
            self.assertEqual(len(redundancy_routers['routers']), 0)

        with self.router() as r, self.port() as p:
            no_subnet = {'id': None, 'network_id': None}
            self._ha_router_port_test(no_subnet, r['router'], p['port'],
                                      None, _disable_ha_tests)

    def test_ha_router_disable_ha_non_admin_succeeds(self):
        def _disable_ha_tests(redundancy_routers):
            body = {'router': {ha.ENABLED: False}}
            updated_router = self._update(
                'routers', r['router']['id'], body,
                neutron_context=bc.context.Context('', tenant_id))
            self._verify_ha_settings(updated_router['router'],
                                     self._get_ha_defaults(ha_enabled=False))
            # verify that the redundancy routers are indeed gone
            params = "&".join(["id=%s" % rr['id'] for rr in
                               redundancy_routers])
            redundancy_routers = self._list('routers', query_params=params)
            self.assertEqual(len(redundancy_routers['routers']), 0)

        tenant_id = _uuid()
        with self.network(tenant_id=tenant_id) as n_external:
            res = self._create_subnet(self.fmt, n_external['network']['id'],
                                      cidr='10.0.1.0/24', tenant_id=tenant_id)
            s = self.deserialize(self.fmt, res)
            self._set_net_external(s['subnet']['network_id'])
            with self.router(
                    external_gateway_info={
                        'network_id': s['subnet']['network_id']},
                    tenant_id=tenant_id) as r:
                with self.port(tenant_id=tenant_id) as p:
                    self._ha_router_port_test(s['subnet'], r['router'],
                                              p['port'], None,
                                              _disable_ha_tests)

    def test_non_gw_ha_router_disable_ha_non_admin_succeeds(self):
        def _disable_ha_tests(redundancy_routers):
            body = {'router': {ha.ENABLED: False}}
            updated_router = self._update(
                'routers', r['router']['id'], body,
                neutron_context=bc.context.Context('', tenant_id))
            self._verify_ha_settings(updated_router['router'],
                                     self._get_ha_defaults(ha_enabled=False))
            # verify that the redundancy routers are indeed gone
            params = "&".join(["id=%s" % rr['id'] for rr in
                               redundancy_routers])
            redundancy_routers = self._list('routers', query_params=params)
            self.assertEqual(len(redundancy_routers['routers']), 0)

        tenant_id = _uuid()
        with self.router(tenant_id=tenant_id) as r:
            with self.port(tenant_id=tenant_id) as p:
                no_subnet = {'id': None, 'network_id': None}
                self._ha_router_port_test(no_subnet, r['router'], p['port'],
                                          None, _disable_ha_tests)

    def test_ha_router_remove_gateway_succeeds(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            subnet = s['subnet']
            self._set_net_external(subnet['network_id'])
            with self.router(external_gateway_info={
                    'network_id': subnet['network_id']}) as r:
                router = r['router']
                # verify router visible to user
                ha_spec = self._get_ha_defaults()
                self._verify_ha_settings(router, ha_spec)
                self._verify_router_ports(router['id'], subnet['network_id'],
                                          subnet['id'])
                # verify redundancy router ports
                for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                    self._verify_router_ports(
                        rr_info['id'], subnet['network_id'], subnet['id'])
                body = {'router': {'external_gateway_info': None}}
                r_after = self._update('routers', router['id'], body)
                self._verify_ha_settings(r_after['router'], ha_spec)
                self._verify_router_ports(router['id'])
                # verify redundancy router ports
                for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                    self._verify_router_ports(rr_info['id'])

    def test_ha_non_gw_router_add_gateway_succeeds(self):
        with self.router() as r:
            router = r['router']
            # verify router visible to user
            ha_spec = self._get_ha_defaults()
            self._verify_ha_settings(router, ha_spec)
            self._verify_router_ports(router['id'])
            # verify redundancy router ports
            for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                self._verify_router_ports(rr_info['id'])
            with self.subnet(cidr='10.0.1.0/24') as s:
                subnet = s['subnet']
                self._set_net_external(subnet['network_id'])
                body = {'router': {'external_gateway_info': {
                    'network_id': subnet['network_id']}}}
                r_after = self._update('routers', router['id'], body)
                self._verify_ha_settings(r_after['router'], ha_spec)
                self._verify_router_ports(router['id'], subnet['network_id'],
                                          subnet['id'])
                # verify redundancy router ports
                for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                    self._verify_router_ports(
                        rr_info['id'], subnet['network_id'], subnet['id'])

    def test_ha_update_admin_state_up(self):
        with self.router() as r:
            router = r['router']
            # verify router visible to user
            ha_spec = self._get_ha_defaults()
            self._verify_ha_settings(router, ha_spec)
            # verify initial state of admin_state_up is True
            self.assertEqual(True, router['admin_state_up'])
            for rr_info in router[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                rr = self._show('routers', rr_info['id'])
                self.assertEqual(True, rr['router']['admin_state_up'])

            # update admin_state_up to False
            body = {'router': {'admin_state_up': False}}
            r_after = self._update('routers', router['id'], body)

            # verify update to router visible to user and the backup routers
            self.assertEqual(False, r_after['router']['admin_state_up'])
            for rr_info in (
                r_after['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS]):
                rr = self._show('routers', rr_info['id'])
                self.assertEqual(False, rr['router']['admin_state_up'])

    def _test_enable_ha(self, subnet, router, port, set_ha_details=True,
                        neutron_context=None):
        if port['network_id']:
            body = self._router_interface_action('add', router['id'], None,
                                                 port['id'])
            self.assertIn('port_id', body)
            self.assertEqual(body['port_id'], port['id'])
        # verify router visible to user
        ha_disabled_settings = self._get_ha_defaults(
            ha_enabled=False)
        self._verify_ha_settings(router, ha_disabled_settings)
        self._verify_router_ports(router['id'], subnet['network_id'],
                                  subnet['id'], port['network_id'],
                                  port['fixed_ips'][0]['subnet_id'])
        body = {'router': {ha.ENABLED: True}}
        if set_ha_details is True:
            body['router'][ha.DETAILS] = {ha.TYPE: ha.HA_VRRP}
        ha_type = body['router'].get(ha.DETAILS, {}).get(ha.TYPE)
        if neutron_context:
            updated_router = self._update('routers', router['id'], body,
                                          neutron_context=neutron_context)
            self._verify_ha_settings(updated_router['router'],
                                     {ha.ENABLED: True})
        else:
            updated_router = self._update('routers', router['id'], body)
            self._verify_ha_settings(updated_router['router'],
                                     self._get_ha_defaults(ha_type=ha_type))
            ha_d = updated_router['router'][ha.DETAILS]
            redundancy_routers = self._list(
                'routers',
                query_params="&".join(["id=%s" % rr['id'] for rr in
                                       ha_d[ha.REDUNDANCY_ROUTERS]]))
            for rr in redundancy_routers['routers']:
                # redundancy router should have ha disabled
                self._verify_ha_settings(rr, ha_disabled_settings)
                # check that redundancy routers have all ports
                self._verify_router_ports(rr['id'], subnet['network_id'],
                                          subnet['id'], port['network_id'],
                                          port['fixed_ips'][0]['subnet_id'])
        if port['network_id']:
            # clean-up
            self._router_interface_action('remove', router['id'], None,
                                          port['id'])

    def test_enable_ha_on_router_succeeds(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {ha.ENABLED: False,
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            with self.router(arg_list=(ha.ENABLED,), **kwargs) as r,\
                    self.port() as p:
                self._test_enable_ha(s['subnet'], r['router'], p['port'])

    def test_enable_ha_on_non_gw_router_succeeds(self):
        kwargs = {ha.ENABLED: False}
        with self.router(arg_list=(ha.ENABLED,), **kwargs) as r,\
                self.port() as p:
            no_subnet = {'id': None, 'network_id': None}
            self._test_enable_ha(no_subnet, r['router'], p['port'])

    def test_enable_ha_on_router_no_itfcs_succeeds_succeeds(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {ha.ENABLED: False,
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            with self.router(arg_list=(ha.ENABLED,), **kwargs) as r:
                no_port = {'id': None, 'network_id': None,
                           'fixed_ips': [{'subnet_id': None}]}
                self._test_enable_ha(s['subnet'], r['router'], no_port)

    def test_enable_ha_on_non_gw_router_no_itfcs_succeeds(self):
        kwargs = {ha.ENABLED: False}
        with self.router(arg_list=(ha.ENABLED,), **kwargs) as r:
            no_subnet = {'id': None, 'network_id': None}
            no_port = {'id': None, 'network_id': None,
                       'fixed_ips': [{'subnet_id': None}]}
            self._test_enable_ha(no_subnet, r['router'], no_port)

    def test_enable_ha_on_router_succeeds_no_ha_spec(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {ha.ENABLED: False,
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            with self.router(arg_list=(ha.ENABLED,), **kwargs) as r:
                with self.port() as p:
                    self._test_enable_ha(s['subnet'], r['router'], p['port'],
                                         set_ha_details=False)

    def test_enable_ha_on_non_gw_router_succeeds_no_ha_spec(self):
        kwargs = {ha.ENABLED: False}
        with self.router(arg_list=(ha.ENABLED,), **kwargs) as r,\
                self.port() as p:
            no_subnet = {'id': None, 'network_id': None}
            self._test_enable_ha(no_subnet, r['router'], p['port'],
                                 set_ha_details=False)

    def test_enable_ha_on_router_non_admin_succeeds(self):
        tenant_id = _uuid()
        with self.network(tenant_id=tenant_id) as n_external:
            res = self._create_subnet(self.fmt, n_external['network']['id'],
                                      cidr='10.0.1.0/24', tenant_id=tenant_id)
            s = self.deserialize(self.fmt, res)
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {ha.ENABLED: False,
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            with self.router(tenant_id=tenant_id, arg_list=(ha.ENABLED,),
                             **kwargs) as r:
                with self.port(tenant_id=tenant_id) as p:
                    self._test_enable_ha(
                        s['subnet'], r['router'], p['port'], False,
                        neutron_context=bc.context.Context('', tenant_id))

    def test_enable_ha_on_non_gw_router_non_admin_succeeds(self):
        tenant_id = _uuid()
        kwargs = {ha.ENABLED: False}
        with self.router(tenant_id=tenant_id, arg_list=(ha.ENABLED,),
                         **kwargs) as r, self.port(tenant_id=tenant_id) as p:
            no_subnet = {'id': None, 'network_id': None}
            self._test_enable_ha(no_subnet, r['router'], p['port'])

    def test_update_router_ha_settings(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s['subnet']['network_id']}) as r:
                self._verify_ha_settings(r['router'], self._get_ha_defaults())
                ha_settings = self._get_ha_defaults(
                    priority=15, probing_enabled=True, probe_interval=3,
                    probe_target='10.5.5.2')
                ha_spec = copy.deepcopy(ha_settings[ha.DETAILS])
                del ha_spec[ha.TYPE]
                del ha_spec[ha.REDUNDANCY_LEVEL]
                body = {'router': {ha.DETAILS: ha_spec}}
                r_after = self._update('routers', r['router']['id'], body)
                self._verify_ha_settings(r_after['router'], ha_settings)
                r_show = self._show('routers', r['router']['id'])
                self._verify_ha_settings(r_show['router'], ha_settings)

    def test_update_non_gw_router_ha_settings(self):
        with self.router() as r:
            self._verify_ha_settings(r['router'], self._get_ha_defaults())
            ha_settings = self._get_ha_defaults(
                priority=15, probing_enabled=True, probe_interval=3,
                probe_target='10.5.5.2')
            ha_spec = copy.deepcopy(ha_settings[ha.DETAILS])
            del ha_spec[ha.TYPE]
            del ha_spec[ha.REDUNDANCY_LEVEL]
            body = {'router': {ha.DETAILS: ha_spec}}
            r_after = self._update('routers', r['router']['id'], body)
            self._verify_ha_settings(r_after['router'], ha_settings)
            r_show = self._show('routers', r['router']['id'])
            self._verify_ha_settings(r_show['router'], ha_settings)

    def test_update_router_ha_settings_non_admin_fails(self):
        tenant_id = _uuid()
        with self.network(tenant_id=tenant_id) as n_external:
            res = self._create_subnet(self.fmt, n_external['network']['id'],
                                      cidr='10.0.1.0/24', tenant_id=tenant_id)
            s = self.deserialize(self.fmt, res)
            self._set_net_external(s['subnet']['network_id'])
            with self.router(
                    external_gateway_info={
                        'network_id': s['subnet']['network_id']},
                    tenant_id=tenant_id) as r:
                ha_settings = self._get_ha_defaults()
                self._verify_ha_settings(r['router'], ha_settings)
                body = {'router': {ha.DETAILS: {ha.PRIORITY: 15,
                                                ha.PROBE_CONNECTIVITY: True,
                                                ha.PROBE_TARGET: '10.5.5.2',
                                                ha.PROBE_INTERVAL: 3}}}
                self._update('routers', r['router']['id'], body,
                             expected_code=webob.exc.HTTPForbidden.code,
                             neutron_context=bc.context.Context('', tenant_id))
                r_show = self._show('routers', r['router']['id'])
                self._verify_ha_settings(r_show['router'], ha_settings)

    def test_update_non_gw_router_ha_settings_non_admin_fails(self):
        tenant_id = _uuid()
        with self.router(tenant_id=tenant_id) as r:
            ha_settings = self._get_ha_defaults()
            self._verify_ha_settings(r['router'], ha_settings)
            body = {'router': {ha.DETAILS: {ha.PRIORITY: 15,
                                            ha.PROBE_CONNECTIVITY: True,
                                            ha.PROBE_TARGET: '10.5.5.2',
                                            ha.PROBE_INTERVAL: 3}}}
            self._update('routers', r['router']['id'], body,
                         expected_code=webob.exc.HTTPForbidden.code,
                         neutron_context=bc.context.Context('', tenant_id))
            r_show = self._show('routers', r['router']['id'])
            self._verify_ha_settings(r_show['router'], ha_settings)

    def test_update_ha_type_on_router_with_ha_enabled_fails(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s['subnet']['network_id']}) as r:
                ha_settings = self._get_ha_defaults()
                self._verify_ha_settings(r['router'], ha_settings)
                body = {'router': {ha.DETAILS: {ha.TYPE: ha.HA_GLBP}}}
                self._update('routers', r['router']['id'], body,
                             expected_code=webob.exc.HTTPConflict.code)
                r_after = self._show('routers', r['router']['id'])
                self._verify_ha_settings(r_after['router'], ha_settings)

    def test_update_ha_type_on_non_gw_router_with_ha_enabled_fails(self):
        with self.router() as r:
            ha_settings = self._get_ha_defaults()
            self._verify_ha_settings(r['router'], ha_settings)
            body = {'router': {ha.DETAILS: {ha.TYPE: ha.HA_GLBP}}}
            self._update('routers', r['router']['id'], body,
                         expected_code=webob.exc.HTTPConflict.code)
            r_after = self._show('routers', r['router']['id'])
            self._verify_ha_settings(r_after['router'], ha_settings)

    def _test_ha_disabled_cases(self):
        with self.subnet(cidr='10.0.1.0/24') as s:
            self._set_net_external(s['subnet']['network_id'])
            kwargs = {ha.ENABLED: False,
                      l3_const.EXTERNAL_GW_INFO: {'network_id':
                                            s['subnet']['network_id']}}
            with self.router(arg_list=(ha.ENABLED,), **kwargs) as r:
                ha_disabled_settings = self._get_ha_defaults(ha_enabled=False)
                self._verify_ha_settings(r['router'], ha_disabled_settings)
                body = {'router': {ha.ENABLED: True,
                                   ha.DETAILS: {ha.TYPE: ha.HA_VRRP}}}
                self._update('routers', r['router']['id'], body,
                             expected_code=webob.exc.HTTPConflict.code)
                r_after = self._show('routers', r['router']['id'])
                self._verify_ha_settings(r_after['router'],
                                         ha_disabled_settings)

    def test_enable_ha_when_ha_support_disabled_fails(self):
        cfg.CONF.set_override('ha_support_enabled', False, group='ha')
        self._test_ha_disabled_cases()

    def test_enable_ha_with_disabled_ha_type_fails(self):
        cfg.CONF.set_override('disabled_ha_mechanisms', [ha.HA_VRRP],
                              group='ha')
        self._test_ha_disabled_cases()

    def _test_ha_disabled_non_gw_cases(self):
        kwargs = {ha.ENABLED: False}
        with self.router(arg_list=(ha.ENABLED,), **kwargs) as r:
            ha_disabled_settings = self._get_ha_defaults(ha_enabled=False)
            self._verify_ha_settings(r['router'], ha_disabled_settings)
            body = {'router': {ha.ENABLED: True,
                               ha.DETAILS: {ha.TYPE: ha.HA_VRRP}}}
            self._update('routers', r['router']['id'], body,
                         expected_code=webob.exc.HTTPConflict.code)
            r_after = self._show('routers', r['router']['id'])
            self._verify_ha_settings(r_after['router'],
                                     ha_disabled_settings)

    def test_enable_ha_when_ha_support_disabled_fails_non_gw(self):
        cfg.CONF.set_override('ha_support_enabled', False, group='ha')
        self._test_ha_disabled_non_gw_cases()

    def test_enable_ha_with_disabled_ha_type_fails_non_gw(self):
        cfg.CONF.set_override('disabled_ha_mechanisms', [ha.HA_VRRP],
                              group='ha')
        self._test_ha_disabled_non_gw_cases()

    def _test_change_ha_router_redundancy_level(self, new_level=1,
                                                gateway=True):

        def _change_redundancy_tests(redundancy_routers):
            new_ha_settings = self._get_ha_defaults(redundancy_level=new_level,
                                                    ha_type=ha.HA_HSRP,
                                                    probing_enabled=False)
            ha_spec = copy.deepcopy(new_ha_settings)
            del ha_spec[ha.DETAILS][ha.PRIORITY]
            updated_router = self._update('routers', r['router']['id'],
                                          {'router': ha_spec})
            # verify router visible to user
            self._verify_ha_settings(updated_router['router'], new_ha_settings)
            self._verify_router_ports(updated_router['router']['id'],
                                      subnet['network_id'], subnet['id'],
                                      port['network_id'],
                                      port['fixed_ips'][0]['subnet_id'])
            ha_d = updated_router['router'][ha.DETAILS]
            params = "&".join(["id=%s" % rr['id'] for rr in
                               ha_d[ha.REDUNDANCY_ROUTERS]])
            res = self._list('routers', query_params=params)
            new_redundancy_routers = res['routers']
            self.assertEqual(len(new_redundancy_routers), new_level)
            ha_disabled_settings = self._get_ha_defaults(ha_enabled=False)
            for rr in new_redundancy_routers:
                # redundancy router should have ha disabled
                self._verify_ha_settings(rr, ha_disabled_settings)
                # check that redundancy router have all ports
                self._verify_router_ports(
                    rr['id'], subnet['network_id'], subnet['id'],
                    port['network_id'], port['fixed_ips'][0]['subnet_id'])
            # verify that non-deleted redundancy routers are the same
            old_rr_ids = set(rr['id'] for rr in redundancy_routers)
            new_rr_ids = set(rr['id'] for rr in new_redundancy_routers)
            if len(old_rr_ids) < len(new_rr_ids):
                self.assertTrue(old_rr_ids.issubset(new_rr_ids))
            else:
                self.assertTrue(new_rr_ids.issubset(old_rr_ids))

        with self.port() as p:
            port = p['port']
            if gateway is True:
                with self.subnet(cidr='10.0.1.0/24') as s:
                    subnet = s['subnet']
                    self._set_net_external(subnet['network_id'])
                    with self.router(external_gateway_info={
                            'network_id': subnet['network_id']}) as r:
                        self._ha_router_port_test(subnet, r['router'], port,
                                                  None,
                                                  _change_redundancy_tests)
            else:
                with self.router() as r:
                    subnet = {'id': None, 'network_id': None}
                    self._ha_router_port_test(subnet, r['router'], port, None,
                                              _change_redundancy_tests)

    def test_decrease_ha_router_redundancy_level(self):
        self._test_change_ha_router_redundancy_level()

    def test_decrease_non_gw_ha_router_redundancy_level(self):
        self._test_change_ha_router_redundancy_level(gateway=False)

    def test_increase_ha_router_redundancy_level(self):
        self._test_change_ha_router_redundancy_level(new_level=3)

    def test_increase_non_gw_ha_router_redundancy_level(self):
        self._test_change_ha_router_redundancy_level(new_level=3,
                                                     gateway=False)

    # Overloaded test function that needs to be modified to run
    def _test_router_add_interface_subnet(self, router, subnet, msg=None):
        exp_notifications = ['router.create.start',
                             'router.create.end',
                             'network.create.start',
                             'network.create.end',
                             'subnet.create.start',
                             'subnet.create.end',
                             'router.interface.create',
                             'router.interface.delete']
        body = self._router_interface_action('add',
                                             router['router']['id'],
                                             subnet['subnet']['id'],
                                             None)
        self.assertIn('port_id', body, msg)

        # fetch port and confirm device_id
        r_port_id = body['port_id']
        port = self._show('ports', r_port_id)
        self.assertEqual(port['port']['device_id'],
                         router['router']['id'], msg)

        self._router_interface_action('remove',
                                      router['router']['id'],
                                      subnet['subnet']['id'],
                                      None)
        self._show('ports', r_port_id,
                   expected_code=webob.exc.HTTPNotFound.code)

        self.assertEqual(
            set(exp_notifications),
            set(n['event_type'] for n in fake_notifier.NOTIFICATIONS), msg)

        # include redundancy routers ids in allowed ids
        r_ids = {router['router']['id'],
                 router['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][0]['id'],
                 router['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][1]['id']}
        for n in fake_notifier.NOTIFICATIONS:
            if n['event_type'].startswith('router.interface.'):
                payload = n['payload']['router_interface']
                self.assertIn('id', payload)
                # assert by set inclusion
                self.assertIn(payload['id'], r_ids)
                self.assertIn('tenant_id', payload)
                if bc.NEUTRON_VERSION >= bc.NEUTRON_PIKE_VERSION:
                    tid = router['router']['tenant_id']
                else:
                    tid = subnet['subnet']['tenant_id']
                # tolerate subnet tenant deliberately set to '' in the
                # nsx metadata access case
                self.assertIn(payload['tenant_id'], [tid, ''], msg)

    # Overloaded test function that needs to be modified to run
    def test_router_list(self):
        with self.router() as v1, self.router() as v2, self.router() as v3:
            routers = [v1, v2, v3]
            for r in routers[:]:
                for rr in r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                    routers.append({'router': {'id': rr['id']}})
            self._test_list_resources('router', routers)

    def test_router_list_with_sort(self):
        with self.router(name='router1') as router1,\
                self.router(name='router2') as router2,\
                self.router(name='router3') as router3:
            routers = []
            for r in (router3, router2, router1):
                rr_info = r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][0]
                rr1 = self._show('routers', rr_info['id'])
                rr_info = r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][1]
                rr2 = self._show('routers', rr_info['id'])
                if rr1['router']['name'] < rr2['router']['name']:
                    routers.append(rr2)
                    routers.append(rr1)
                else:
                    routers.append(rr1)
                    routers.append(rr2)
                routers.append(r)
            self._test_list_with_sort('router', routers, [('name', 'desc')])

    # Overloaded test function that needs to be modified to run
    def test_router_list_with_pagination(self):
        with self.router(name='router1') as router1,\
                self.router(name='router2') as router2,\
                self.router(name='router3') as router3:
            routers = []
            for r in (router1, router2, router3):
                rr_info = r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][0]
                rr1 = self._show('routers', rr_info['id'])
                rr_info = r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][1]
                rr2 = self._show('routers', rr_info['id'])
                routers.append(r)
                if rr1['router']['name'] > rr2['router']['name']:
                    routers.append(rr2)
                    routers.append(rr1)
                else:
                    routers.append(rr1)
                    routers.append(rr2)
            self._test_list_with_pagination('router', routers, ('name', 'asc'),
                                            3, 4)

    # Overloaded test function that needs to be modified to run
    def test_router_list_with_pagination_reverse(self):
        with self.router(name='router1') as router1,\
                self.router(name='router2') as router2,\
                self.router(name='router3') as router3:
            routers = []
            for r in (router1, router2, router3):
                rr_info = r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][0]
                rr1 = self._show('routers', rr_info['id'])
                rr_info = r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS][1]
                rr2 = self._show('routers', rr_info['id'])
                routers.append(r)
                if rr1['router']['name'] > rr2['router']['name']:
                    routers.append(rr2)
                    routers.append(rr1)
                else:
                    routers.append(rr1)
                    routers.append(rr2)
            self._test_list_with_pagination_reverse('router', routers,
                                                    ('name', 'asc'), 4, 3)

    # Overloaded test function that needs to be modified to run
    def test_floatingip_multi_external_one_internal(self):
        with self.subnet(cidr="10.0.0.0/24") as exs1,\
                self.subnet(cidr="11.0.0.0/24") as exs2,\
                self.subnet(cidr="12.0.0.0/24") as ins1:
            network_ex_id1 = exs1['subnet']['network_id']
            network_ex_id2 = exs2['subnet']['network_id']
            self._set_net_external(network_ex_id1)
            self._set_net_external(network_ex_id2)

            r2i_fixed_ips = [{'ip_address': '12.0.0.2'}]
            with self.router() as r1,\
                    self.router() as r2,\
                    self.port(subnet=ins1,
                              fixed_ips=r2i_fixed_ips) as r2i_port:
                self._add_external_gateway_to_router(
                    r1['router']['id'],
                    network_ex_id1)
                self._router_interface_action('add', r1['router']['id'],
                                              ins1['subnet']['id'],
                                              None)
                self._add_external_gateway_to_router(
                    r2['router']['id'],
                    network_ex_id2)
                self._router_interface_action('add', r2['router']['id'],
                                              None,
                                              r2i_port['port']['id'])

                with self.port(subnet=ins1,
                    # Use 12.0.0.199 instead of 12.0.0.3 so that the IP
                    # address we're requesting is definitely available
                               fixed_ips=[{'ip_address': '12.0.0.199'}]
                               ) as private_port:
                    # Use 10.0.0.199 instead of 10.0.0.3 so that the IP
                    # address we're requesting is definitely available
                    fp1 = self._make_floatingip(self.fmt, network_ex_id1,
                                                private_port['port']['id'],
                                                floating_ip='10.0.0.199')
                    # Use 11.0.0.199 instead of 11.0.0.3 so that the IP
                    # address we're requesting is definitely available
                    fp2 = self._make_floatingip(self.fmt, network_ex_id2,
                                                private_port['port']['id'],
                                                floating_ip='11.0.0.199')
                    self.assertEqual(fp1['floatingip']['router_id'],
                                     r1['router']['id'])
                    self.assertEqual(fp2['floatingip']['router_id'],
                                     r2['router']['id'])

    # Overloaded test function that needs to be modified to run
    def test_router_update_on_external_port(self):
        with self.router() as r:
            with self.subnet(cidr='10.0.1.0/24') as s:
                self._set_net_external(s['subnet']['network_id'])
                self._add_external_gateway_to_router(
                    r['router']['id'],
                    s['subnet']['network_id'])
                body = self._show('routers', r['router']['id'])
                net_id = body['router']['external_gateway_info']['network_id']
                self.assertEqual(net_id, s['subnet']['network_id'])
                port_res = self._list_ports(
                    'json',
                    200,
                    s['subnet']['network_id'],
                    tenant_id=r['router']['tenant_id'],
                    device_owner=bc.constants.DEVICE_OWNER_ROUTER_GW)
                port_list = self.deserialize('json', port_res)
                # Need to change to 3 as there are two ports for the
                # redundancy routers
                self.assertEqual(len(port_list['ports']), 3)

                routes = [{'destination': '135.207.0.0/16',
                           'nexthop': '10.0.1.199'}]

                body = self._update('routers', r['router']['id'],
                                    {'router': {'routes':
                                                routes}})

                body = self._show('routers', r['router']['id'])
                self.assertEqual(body['router']['routes'],
                                 routes)

                # Need to assert that the routes for the redundancy routers
                # have also been updated
                for rr_info in r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                    body = self._show('routers', rr_info['id'])
                    self.assertEqual(body['router']['routes'],
                                     routes)

                self._remove_external_gateway_from_router(
                    r['router']['id'],
                    s['subnet']['network_id'])
                body = self._show('routers', r['router']['id'])
                r_after = body['router']
                gw_info = r_after['external_gateway_info']
                self.assertIsNone(gw_info)
                # Need to assert that the gateways in the redundancy
                # routers have also been updated
                for rr_info in r_after[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                    body = self._show('routers', rr_info['id'])
                    gw_info = body['router']['external_gateway_info']
                    self.assertIsNone(gw_info)

    # Overloaded test function that needs to be modified to run
    def test_floatingip_same_external_and_internal(self):
        # Select router with subnet's gateway_ip for floatingip when
        # routers connected to same subnet and external network.
        with self.subnet(cidr="10.0.0.0/24") as exs, \
                self.subnet(cidr="12.0.0.0/24", gateway_ip="12.0.0.50") as ins:
            network_ex_id = exs['subnet']['network_id']
            self._set_net_external(network_ex_id)

            r2i_fixed_ips = [{'ip_address': '12.0.0.2'}]
            with self.router() as r1, \
                    self.router() as r2, \
                    self.port(subnet=ins,
                              fixed_ips=r2i_fixed_ips) as r2i_port:
                self._add_external_gateway_to_router(
                    r1['router']['id'],
                    network_ex_id)
                self._router_interface_action('add', r2['router']['id'],
                                              None,
                                              r2i_port['port']['id'])
                self._router_interface_action('add', r1['router']['id'],
                                              ins['subnet']['id'],
                                              None)
                self._add_external_gateway_to_router(
                    r2['router']['id'],
                    network_ex_id)
                # Use 12.0.0.199 instead of 12.0.0.14 so that the IP
                # address we're requesting is definitely available
                with self.port(subnet=ins,
                               fixed_ips=[{'ip_address': '12.0.0.199'}]
                               ) as private_port:
                    fp = self._make_floatingip(self.fmt, network_ex_id,
                                               private_port['port']['id'])
                    self.assertEqual(r1['router']['id'],
                                     fp['floatingip']['router_id'])

    def test_router_update_change_external_gateway_and_routes(self):
        with self.router() as router:
            r = router['router']
            with self.subnet(cidr='12.0.1.0/24') as s_priv,\
                    self.subnet(cidr='10.0.1.0/24') as s_ext_1:
                self._set_net_external(s_ext_1['subnet']['network_id'])
                self._add_external_gateway_to_router(r['id'],
                    s_ext_1['subnet']['network_id'])
                body = self._show('routers', r['id'])
                net_id = body['router']['external_gateway_info']['network_id']
                self.assertEqual(net_id, s_ext_1['subnet']['network_id'])
                self._router_interface_action('add', r['id'],
                                              s_priv['subnet']['id'], None)
                port_res = self._list_ports(
                    'json',
                    200,
                    s_ext_1['subnet']['network_id'],
                    tenant_id=r['tenant_id'],
                    device_owner=bc.constants.DEVICE_OWNER_ROUTER_GW)
                gw_port_list = self.deserialize('json', port_res)
                # There are one gw port on the user visible router and two gw
                # ports on the redundancy routers
                self.assertEqual(len(gw_port_list['ports']), 3)
                port_res = self._list_ports(
                    'json',
                    200,
                    s_priv['subnet']['network_id'],
                    tenant_id=r['tenant_id'],
                    device_owner=bc.constants.DEVICE_OWNER_ROUTER_INTF)
                port_list = self.deserialize('json', port_res)
                # There are one interface VIP port and one extra port (for
                # interface ip) on the user visible router and two ports on
                # the redundancy routers (for their interface ips)
                self.assertEqual(len(port_list['ports']), 4)
                with self.subnet(cidr='11.0.1.0/24') as s_ext_2:
                    self._set_net_external(s_ext_2['subnet']['network_id'])
                    new_ext_gw = {
                        'network_id': s_ext_2['subnet']['network_id']}
                    routes = [{'destination': '135.207.0.0/16',
                               'nexthop': '12.0.1.55'}]
                    # change gateway and a route using that gateway
                    body = self._update('routers', r['id'],
                                        {'router': {
                                            EXTERNAL_GW_INFO: new_ext_gw,
                                            'routes': routes}})
                    body = self._show('routers', r['id'])
                    self.assertEqual(body['router']['routes'],
                                     routes)

                    # Need to assert that the routes for the redundancy routers
                    # have also been updated
                    self.assertEqual(
                        len(r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]), 2)
                    for rr_info in r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                        body = self._show('routers', rr_info['id'])
                        self.assertEqual(body['router']['routes'],
                                         routes)

                    self._remove_external_gateway_from_router(
                        r['id'], s_ext_2['subnet']['network_id'])
                    body = self._show('routers', r['id'])
                    r_after = body['router']
                    gw_info = r_after['external_gateway_info']
                    self.assertIsNone(gw_info)
                    # Need to assert that the gateways in the redundancy
                    # routers have also been updated
                    for rr_info in r_after[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                        body = self._show('routers', rr_info['id'])
                        gw_info = body['router']['external_gateway_info']
                        self.assertIsNone(gw_info)
                    routes = []
                    # Remove the route
                    body = self._update('routers', r['id'],
                                        {'router': {'routes': routes}})
                    body = self._show('routers', r['id'])
                    self.assertEqual(body['router']['routes'],
                                     routes)

                    # Need to assert that the routes for the redundancy routers
                    # have also been updated
                    self.assertEqual(
                        len(r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]), 2)
                    for rr_info in r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]:
                        body = self._show('routers', rr_info['id'])
                        self.assertEqual(body['router']['routes'],
                                         routes)

                    self._router_interface_action('remove', r['id'],
                                                  s_priv['subnet']['id'], None)

    def test_redundancy_router_routes_is_from_user_visible_router(self):

        routes = [{'destination': '135.207.0.0/16', 'nexthop': '10.0.1.199'},
                  {'destination': '12.0.0.0/8', 'nexthop': '10.0.1.200'},
                  {'destination': '141.212.0.0/16', 'nexthop': '10.0.1.201'}]
        with self.router() as router,\
                self.subnet(cidr='10.0.1.0/24') as subnet,\
                self.port(subnet=subnet) as port:
            r = router['router']
            p = port['port']
            updated_r = self._routes_update_prepare(r['id'], None, p['id'],
                                                    routes)['router']
            params = "&".join(["id=%s" % rr['id'] for rr in
                               r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]])
            routers = self._list('routers', query_params=params)['routers']
            routers.append(updated_r)
            correct_routes = _sort_routes(routes)
            for router in routers:
                self.assertEqual(_sort_routes(router['routes']),
                                 correct_routes)
            self._routes_update_cleanup(p['id'], None, r['id'], [])

    def _rr_routes_update_prepare(self, router_id, subnet_id, port_id,
                                  routes_router_id, routes, skip_add=False):
        if not skip_add:
            self._router_interface_action('add', router_id, subnet_id, port_id)
        self._update('routers', routes_router_id, {'router': {
            'routes': routes}})
        return self._show('routers', routes_router_id)

    def _rr_routes_update_cleanup(self, port_id, subnet_id, router_id,
                                  routes_router_id, routes):
        self._update('routers', routes_router_id, {'router': {
            'routes': routes}})
        self._router_interface_action('remove', router_id, subnet_id, port_id)

    def test_redundancy_router_routes_includes_user_visible_router(self):
        routes1 = [{'destination': '135.207.0.0/16', 'nexthop': '10.0.1.199'},
                   {'destination': '12.0.0.0/8', 'nexthop': '10.0.1.200'},
                   {'destination': '141.212.0.0/16', 'nexthop': '10.0.1.201'}]
        with self.router() as router,\
                self.subnet(cidr='10.0.1.0/24') as subnet1,\
                self.port(subnet=subnet1) as port1:
            r = router['router']
            p1 = port1['port']
            updated_r = self._routes_update_prepare(r['id'], None, p1['id'],
                                                    routes1)['router']
            params = "id=%s" % r[ha.DETAILS][ha.REDUNDANCY_ROUTERS][1]['id']
            routers = self._list('routers', query_params=params)['routers']
            routers.append(updated_r)
            correct_routes1 = _sort_routes(routes1)
            for router in routers:
                self.assertEqual(_sort_routes(router['routes']),
                                 correct_routes1)
            self._routes_update_cleanup(p1['id'], None, r['id'], [])

    def test_router_update_change_name_changes_redundancy_routers(self):
        with self.router() as router:
            r = router['router']
            newName = 'routerOne'
            params = "&".join(["id=%s" % rr['id'] for rr in
                               r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]])
            r_routers = self._list('routers', query_params=params)['routers']
            rr_name_start = 'router1' + ha_db.REDUNDANCY_ROUTER_SUFFIX
            for rr in r_routers:
                self.assertTrue(rr['name'].startswith(rr_name_start))
            r_updated = self._update('routers', r['id'],
                                     {'router': {'name': newName}})['router']
            self.assertEqual(newName, r_updated['name'])
            rr_name_start = newName + ha_db.REDUNDANCY_ROUTER_SUFFIX
            r_routers = self._list('routers', query_params=params)['routers']
            for rr in r_routers:
                self.assertTrue(rr['name'].startswith(rr_name_start))

    def test_name_change_of_redundancy_router_does_not_create_another_one(
            self):
        with self.router() as router:
            r = router['router']
            params = "&".join(["id=%s" % rr['id'] for rr in
                               r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]])
            r_routers = self._list('routers', query_params=params)['routers']
            self.assertEqual(2, len(r_routers))
            newName = 'routerOne'
            admin_ctx = bc.context.get_admin_context()
            # We call the plugin function directly rather than calling via the
            # REST API as the latter would be rejected since it's a router
            # managed by the L3 router service plugin.
            r_updated = self.l3_plugin.update_router(
                admin_ctx, r_routers[0]['id'], {'router': {'name': newName}})
            self.assertEqual(newName, r_updated['name'])
            routers = self._list('routers')['routers']
            # should have no new routers
            self.assertEqual(3, len(routers))
            r_ids = [rr['id'] for rr in r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]]
            r_ids.append(r['id'])
            for r in routers:
                self.assertIn(r['id'], r_ids)
                r_ids.remove(r['id'])

    def test__notify_subnetpool_address_scope_update(self):
        l3_plugin = bc.get_plugin(bc.constants.L3)

        tenant_id = _uuid()
        with mock.patch.object(
                l3_plugin, 'notify_routers_updated') as chk_method, \
                self.subnetpool(prefixes=['10.0.0.0/24'],
                                admin=True, name='sp',
                                tenant_id=tenant_id) as subnetpool, \
                self.router(tenant_id=tenant_id) as router, \
                self.network(tenant_id=tenant_id) as network:
            subnetpool_id = subnetpool['subnetpool']['id']
            data = {'subnet': {
                    'network_id': network['network']['id'],
                    'subnetpool_id': subnetpool_id,
                    'prefixlen': 24,
                    'ip_version': 4,
                    'tenant_id': tenant_id}}
            req = self.new_create_request('subnets', data)
            subnet = self.deserialize(self.fmt, req.get_response(self.api))

            admin_ctx = bc.context.get_admin_context()
            l3_plugin.add_router_interface(
                admin_ctx,
                router['router']['id'], {'subnet_id': subnet['subnet']['id']})
            if bc.NEUTRON_VERSION >= bc.NEUTRON_PIKE_VERSION:
                (l3_db.L3RpcNotifierMixin.
                    _notify_subnetpool_address_scope_update(
                        mock.ANY, mock.ANY, mock.ANY, context=admin_ctx,
                        subnetpool_id=subnetpool_id))
            else:
                l3_db._notify_subnetpool_address_scope_update(
                    mock.ANY, mock.ANY, mock.ANY,
                    context=admin_ctx, subnetpool_id=subnetpool_id)
            args, kwargs = chk_method.call_args
            self.assertEqual(admin_ctx, args[0])
            self.assertIn(router['router']['id'], args[1])

    @unittest.skip('Gateway network should have a subnet')
    def test_router_add_gateway_no_subnet(self):
        pass


class L3AgentHARouterApplianceTestCase(
        test_l3_router_appliance_plugin.L3AgentRouterApplianceTestCase):

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        if ext_mgr is None:
            ext_mgr = TestHAL3RouterApplianceExtensionManager()
        super(L3AgentHARouterApplianceTestCase, self).setUp(
            l3_plugin=l3_plugin, ext_mgr=ext_mgr)


class L3CfgAgentHARouterApplianceTestCase(
    HAL3RouterTestsMixin,
        test_l3_router_appliance_plugin.L3CfgAgentRouterApplianceTestCase):

    _is_ha_tests = True

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        if l3_plugin is None:
            l3_plugin = L3_PLUGIN_KLASS
        if ext_mgr is None:
            ext_mgr = TestHAL3RouterApplianceExtensionManager()
        cfg.CONF.set_override('default_ha_redundancy_level', 2, group='ha')

        super(L3CfgAgentHARouterApplianceTestCase, self).setUp(
            l3_plugin=l3_plugin, ext_mgr=ext_mgr)
        self.orig_get_sync_data = self.l3_plugin.get_sync_data
        self.l3_plugin.get_sync_data = self.l3_plugin.get_sync_data_ext

    def tearDown(self):
        self.l3_plugin.get_sync_data = self.orig_get_sync_data
        super(L3CfgAgentHARouterApplianceTestCase, self).tearDown()

    # Overloaded test function that needs to be modified to run
    def test_l3_agent_routers_query_interfaces(self):
        with self.router() as r:
            with self.port() as p:
                self._router_interface_action('add',
                                              r['router']['id'],
                                              None,
                                              p['port']['id'])

                routers = self.l3_plugin.get_sync_data(
                    bc.context.get_admin_context(), None)
                self.assertEqual(3, len(routers))
                for router in routers:
                    interfaces = router[bc.constants.INTERFACE_KEY]
                    self.assertEqual(1, len(interfaces))
                    subnets = interfaces[0]['subnets']
                    self.assertEqual(1, len(subnets))
                    subnet_id = subnets[0]['id']
                    wanted_subnetid = p['port']['fixed_ips'][0]['subnet_id']
                    self.assertEqual(wanted_subnetid, subnet_id)

    # Overloaded test function that needs to be modified to run
    def test_l3_agent_routers_query_ignore_interfaces_with_moreThanOneIp(self):
        with self.router() as r:
            with self.subnet(cidr='9.0.1.0/24') as subnet:
                with self.port(subnet=subnet,
                               fixed_ips=[{'ip_address': '9.0.1.3'}]) as p:
                    self._router_interface_action('add',
                                                  r['router']['id'],
                                                  None,
                                                  p['port']['id'])
                    # IP addresses 9.0.1.4 and 9.0.1.5 are consumed by the
                    # redundancy routers to we need to use other addresses.
                    port = {'port': {'fixed_ips':
                                     [{'ip_address': '9.0.1.199',
                                       'subnet_id': subnet['subnet']['id']},
                                      {'ip_address': '9.0.1.200',
                                       'subnet_id': subnet['subnet']['id']}]}}
                    ctx = bc.context.get_admin_context()
                    self.core_plugin.update_port(ctx, p['port']['id'], port)
                    routers = self.l3_plugin.get_sync_data(ctx, None)
                    # One user visible router and two redundancy routers
                    self.assertEqual(3, len(routers))
                    # One interface for the user visible router and one each
                    # for the two redundancy routers
                    for r in routers:
                        interfaces = r.get(bc.constants.INTERFACE_KEY, [])
                        self.assertEqual(1, len(interfaces))

    # Overloaded test function that needs to be modified to run
    @unittest.skipIf(bc.NEUTRON_VERSION < bc.NEUTRON_NEWTON_VERSION,
                     "Test not applicable prior to Newton")
    def test_router_delete_precommit_event(self):
        deleted = set()
        auditor = lambda *a, **k: deleted.add(k['router_id'])
        registry.subscribe(auditor, resources.ROUTER, events.PRECOMMIT_DELETE)
        with self.router() as r:
            self._delete('routers', r['router']['id'])
        r_ids = {rr['id'] for rr in
                 r['router'][ha.DETAILS][ha.REDUNDANCY_ROUTERS]}
        r_ids.add(r['router']['id'])
        self.assertEqual(r_ids, deleted)

    def _test_notify_op_agent(self, target_func, *args):
        kargs = [item for item in args]
        kargs.append(self._l3_cfg_agent_mock)
        target_func(*kargs)

    def _routes_update_prepare(self, router_id, subnet_id, port_id,
                               routes_router_id, routes, skip_add=False):
        if not skip_add:
            self._router_interface_action('add', router_id, subnet_id, port_id)
        self._update('routers', routes_router_id, {'router': {
            'routes': routes}})
        return self._show('routers', routes_router_id)

    def _routes_update_cleanup(self, port_id, subnet_id, router_id,
                               routes_router_id, routes):
        self._update('routers', routes_router_id, {'router': {
            'routes': routes}})
        self._router_interface_action('remove', router_id, subnet_id, port_id)

    def _test_ha_routes_op_cfg_agent(self, notifyApi):
        routes = [{'destination': '135.207.0.0/16', 'nexthop': '10.0.1.199'},
                  {'destination': '12.0.0.0/8', 'nexthop': '10.0.1.200'},
                  {'destination': '141.212.0.0/16', 'nexthop': '10.0.1.201'}]
        with self.router() as router,\
                self.subnet(cidr='10.0.1.0/24') as subnet,\
                self.port(subnet=subnet) as port:
            r = router['router']
            p = port['port']
            self._routes_update_prepare(r['id'], None, p['id'], r['id'],
                                        routes)
            rr_ids = [
                rr['id'] for rr in r['cisco_ha:details']['redundancy_routers']]

            # notifications for interface add etc are sent first which we
            # don't care about in this test
            # Then when the update for routes is triggered,
            # notification is sent for redundant routers first
            call_args_rr_notif = (
                notifyApi.routers_updated.call_args_list[3][0][1])
            rr_ids_in_notification = [rr['id'] for rr in call_args_rr_notif]
            # Check if the notification includes all redundant routers
            self.assertEqual(sorted(rr_ids_in_notification),
                sorted(rr_ids))
            # notification is sent for user visible router next
            call_args_user_r_notif = (
                notifyApi.routers_updated.call_args_list[4][0][1])
            r_id_in_notification = call_args_user_r_notif[0]['id']
            self.assertEqual(r_id_in_notification, r['id'])
            self._routes_update_cleanup(p['id'], None, r['id'], r['id'], [])

    def test_ha_routes_op_cfg_agent(self):
        self._test_notify_op_agent(self._test_ha_routes_op_cfg_agent)

    def test_l3_cfg_agent_query_ha_rdcy_router_routes_is_from_user_vsbl_router(
            self):
        routes = [{'destination': '135.207.0.0/16', 'nexthop': '10.0.1.199'},
                  {'destination': '12.0.0.0/8', 'nexthop': '10.0.1.200'},
                  {'destination': '141.212.0.0/16', 'nexthop': '10.0.1.201'}]
        with self.router() as router,\
                self.subnet(cidr='10.0.1.0/24') as subnet,\
                self.port(subnet=subnet) as port:
            r = router['router']
            p = port['port']
            self._routes_update_prepare(r['id'], None, p['id'], r['id'],
                                        routes)
            router_ids = [rr['id']
                          for rr in r[ha.DETAILS][ha.REDUNDANCY_ROUTERS]]
            router_ids.append(r['id'])
            e_context = bc.context.get_admin_context()
            routers = self.l3_plugin.get_sync_data_ext(e_context, router_ids)
            self.assertEqual(len(router_ids), len(routers))
            correct_routes = _sort_routes(routes)
            for router in routers:
                self.assertEqual(_sort_routes(router['routes']),
                                 correct_routes)
            self._routes_update_cleanup(p['id'], None, r['id'], r['id'], [])

    def test_l3_cfg_agent_query_ha_rdcy_router_routes_include_user_vsbl_router(
            self):
        routes1 = [{'destination': '135.207.0.0/16', 'nexthop': '10.0.1.199'},
                   {'destination': '12.0.0.0/8', 'nexthop': '10.0.1.200'},
                   {'destination': '141.212.0.0/16', 'nexthop': '10.0.1.201'}]
        with self.router() as router,\
                self.subnet(cidr='10.0.1.0/24') as subnet1,\
                self.port(subnet=subnet1) as port1:
            r = router['router']
            p1 = port1['port']
            self._routes_update_prepare(r['id'], None, p1['id'], r['id'],
                                        routes1)
            router_ids = [r['id'],
                          r[ha.DETAILS][ha.REDUNDANCY_ROUTERS][1]['id']]
            e_context = bc.context.get_admin_context()
            routers = self.l3_plugin.get_sync_data_ext(e_context, router_ids)
            self.assertEqual(len(router_ids), len(routers))
            correct_routes1 = _sort_routes(routes1)
            for router in routers:
                self.assertEqual(_sort_routes(router['routes']),
                                 correct_routes1)
            self._routes_update_cleanup(p1['id'], None, r['id'], r['id'], [])

    def test_l3_cfg_agent_query_ha_router_with_fips(self):
        with self.subnet(cidr='10.0.1.0/24') as s_ext,\
                self.subnet(cidr='10.0.2.0/24') as s1,\
                self.subnet(cidr='10.0.3.0/24') as s2:
            self._set_net_external(s_ext['subnet']['network_id'])
            with self.router(external_gateway_info={
                    'network_id': s_ext['subnet']['network_id'],
                    'external_fixed_ips': [{'ip_address': '10.0.1.2'}]}) as r:
                ipspec1 = [{'subnet_id': s1['subnet']['id'],
                            'ip_address': s1['subnet']['gateway_ip']}]
                ipspec2 = [{'subnet_id': s2['subnet']['id'],
                            'ip_address': s2['subnet']['gateway_ip']}]
                with self.port(subnet=s1, fixed_ips=ipspec1) as p1,\
                        self.port(subnet=s1) as private_p1,\
                        self.port(subnet=s2, fixed_ips=ipspec2) as p2,\
                        self.port(subnet=s2) as private_p2:
                    self._router_interface_action(
                        'add', r['router']['id'], None, p1['port']['id'])
                    self._router_interface_action(
                        'add', r['router']['id'], None, p2['port']['id'])
                    fip1 = self._make_floatingip(
                        self.fmt,
                        s_ext['subnet']['network_id'],
                        port_id=private_p1['port']['id'])
                    fip1['floatingip']['fixed_ip_address_scope'] = None
                    fip2 = self._make_floatingip(
                        self.fmt,
                        s_ext['subnet']['network_id'],
                        port_id=private_p2['port']['id'])
                    fip2['floatingip']['fixed_ip_address_scope'] = None
                    fips_dict = {fip1['floatingip']['id']: fip1['floatingip'],
                                 fip2['floatingip']['id']: fip2['floatingip']}
                    e_context = bc.context.get_admin_context()
                    query_params = """fixed_ips=ip_address%%3D%s""".strip() % (
                                   '10.0.1.2')
                    gw_port = self._list('ports',
                                         query_params=query_params)['ports'][0]
                    ports = {gw_port['id']: gw_port,
                             p1['port']['id']: p1['port'],
                             p2['port']['id']: p2['port']}
                    ha_groups_dict = {}
                    ha_settings = self._get_ha_defaults()
                    routers = self._validate_router_sync_data(
                        e_context, [r['router']['id']], s_ext, ports,
                        ha_settings, ha_groups_dict, fips_dict)
                    rr_ids = [rr['id'] for rr in routers[0][ha.DETAILS][
                        ha.REDUNDANCY_ROUTERS]]
                    del fip1['floatingip']['fixed_ip_address_scope']
                    del fip2['floatingip']['fixed_ip_address_scope']
                    # redundancy routers should here have same ha settings
                    # as the user visible routers since the l3 cfg agent
                    # needs that information to configure the redundancy
                    # router
                    self._validate_router_sync_data(
                        e_context, rr_ids, s_ext, ports, ha_settings,
                        ha_groups_dict, fips_dict)
                    # clean-up
                    self._delete('floatingips', fip2['floatingip']['id'])
                    self._delete('floatingips', fip1['floatingip']['id'])
                    self._router_interface_action('remove', r['router']['id'],
                                                  None, p2['port']['id'])
                    self._router_interface_action('remove', r['router']['id'],
                                                  None, p1['port']['id'])

    def _validate_router_sync_data(self, context, router_ids, external_subnet,
                                   ports, ha_settings, ha_groups_dict,
                                   fips_dict):
            routers = self.l3_plugin.get_sync_data_ext(context, router_ids)
            self.assertEqual(len(router_ids), len(routers))
            for r in routers:
                self.assertEqual(external_subnet['subnet']['id'],
                                 r['gw_port']['subnets'][0]['id'])
                # redundancy routers should here have same ha settings
                # as the user visible routers since the l3 cfg agent
                # needs that information to configure the redundancy
                # router
                self._verify_ha_settings(r, ha_settings)
                # the id of this redundancy router should be in the
                # list of redundancy routers
                rr_ids = [rr['id'] for rr in r[ha.DETAILS][
                    ha.REDUNDANCY_ROUTERS]]
                r_fips = r.get(bc.constants.FLOATINGIP_KEY, [])
                self.assertEqual(len(r_fips), len(fips_dict))
                for r_fip in r_fips:
                    self.assertEqual(r_fip, fips_dict[r_fip['id']])
                if ha_groups_dict:
                    # the id of a redundancy router should be in the
                    # list of redundancy routers
                    self.assertIn(r['id'], rr_ids)
                else:
                    # but not the id of a user visible router
                    self.assertNotIn(r['id'], rr_ids)
                # adding the router gw port to the list of internal router port
                # since we want to run the identical tests for all of them
                r[bc.constants.INTERFACE_KEY].append(r['gw_port'])
                self._validate_router_interface_ha_info(
                    ports, r[bc.constants.INTERFACE_KEY],
                    ha_groups_dict)
            return routers

    def _validate_router_interface_ha_info(self, ports_dict, interfaces,
                                           ha_groups_dict):
        self.assertEqual(len(ports_dict), len(interfaces))
        assemble_groups = len(ha_groups_dict) == 0
        for i in interfaces:
            ha_info = i[ha.HA_INFO]
            self.assertIsNotNone(ha_info)
            if assemble_groups:
                ha_groups_dict[ha_info[ha_db.HA_PORT]['id']] = ha_info[
                    ha_db.HA_GROUP]
            else:
                ha_port_id = ha_info[ha_db.HA_PORT]['id']
                self.assertIsNotNone(ports_dict.get(ha_port_id))
                self.assertEqual(ha_info[ha_db.HA_GROUP],
                                 ha_groups_dict[ha_port_id])

    def _test_router_gateway_op_agent(self, notifyApi):
        with self.router() as r:
            with self.subnet() as s:
                self._set_net_external(s['subnet']['network_id'])
                self._add_external_gateway_to_router(
                    r['router']['id'],
                    s['subnet']['network_id'])
                self._remove_external_gateway_from_router(
                    r['router']['id'],
                    s['subnet']['network_id'])
                # one user visible router and two redundancy routers (for
                # which a single update contains both redundancy routers), so
                # 1 + 1 add gateway + 1 + 1 remove gateway = 4 uodates
                self.assertEqual(
                    4, notifyApi.routers_updated.call_count)

    def _test_interfaces_op_agent(self, r, notifyApi):
        with self.port() as p:
            self._router_interface_action('add',
                                          r['router']['id'],
                                          None,
                                          p['port']['id'])
            # clean-up
            self._router_interface_action('remove',
                                          r['router']['id'],
                                          None,
                                          p['port']['id'])
        # one user visible router and two redundancy routers, so
        # 3 x add interface + 3 x remove interface
        self.assertEqual(6, notifyApi.routers_updated.call_count)

    def _test_floatingips_op_agent(self, notifyApi):
        with self.floatingip_with_assoc():
            pass
        # 2 x add gateway (one for user visible router), one for redundancy
        # routers
        # 3 x add interface (one for each router),
        # 1 x creation of floatingip (with 3 routers included),
        # 1 x deletion of floatingip (with 3 routers included)
        self.assertEqual(7, notifyApi.routers_updated.call_count)

    def _validate_ha_fip_ops(self, notifyApi, routers, first_operation):
        # 2 x add gateway (one for user visible router), one for redundancy
        # routers
        # 3 x add interface (one for each router),
        # 1 x update of floatingip (with 3 routers included),
        # 1 x deletion of floatingip (with 3 routers included)
        notify_call_1 = notifyApi.routers_updated.mock_calls[5]
        self.assertEqual(notify_call_1[1][2], first_operation)
        r_ids = {r['id'] for r in notify_call_1[1][1]}
        for r in routers:
            self.assertIn(r['id'], r_ids)
            r_ids.remove(r['id'])
        self.assertEqual(len(r_ids), 0)
        delete_call = notifyApi.routers_updated.mock_calls[6]
        self.assertEqual(delete_call[1][2], 'delete_floatingip')
        r_ids = {r['id'] for r in delete_call[1][1]}
        for r in routers:
            self.assertIn(r['id'], r_ids)
            r_ids.remove(r['id'])
        self.assertEqual(len(r_ids), 0)
        self.assertEqual(7, notifyApi.routers_updated.call_count)

    def _test_ha_floatingips_op_cfg_agent(self, notifyApi):
        with self.floatingip_with_assoc():
            routers = self._list('routers')['routers']
        self._validate_ha_fip_ops(notifyApi, routers, 'create_floatingip')

    def test_ha_floatingips_op_cfg_agent(self):
        self._test_notify_op_agent(self._test_ha_floatingips_op_cfg_agent)

    def _test_ha_floatingip_update_cfg_agent(self, notifyApi):
        with self.subnet() as private_sub:
            with self.port(private_sub) as p_port:
                private_port = p_port['port']
                with self.floatingip_no_assoc(private_sub) as fl_ip:
                    fip = fl_ip['floatingip']
                    routers = self._list('routers')['routers']
                    fip_spec = {'floatingip': {'port_id': private_port['id']}}
                    self._update('floatingips', fip['id'], fip_spec)
        self._validate_ha_fip_ops(notifyApi, routers, 'update_floatingip')

    def test_ha_floatingip_update_cfg_agent(self):
        self._test_notify_op_agent(self._test_ha_floatingip_update_cfg_agent)

    def test_populate_port_ha_information_no_port(self):

        with self.router(arg_list=(ha.ENABLED,)) as router:
            r = router['router']
            with self.port() as port:
                p = port['port']
                body = self._router_interface_action('add', r['id'], None,
                                                     p['id'])
                self.assertIn('port_id', body)
                self.assertEqual(body['port_id'], p['id'])
                adm_ctx = bc.context.get_admin_context()
                with mock.patch(
                    'networking_cisco.plugins.cisco.db.l3.ha_db.HA_db_mixin.'
                    '_populate_port_ha_information') as mock_port_ha:
                    mock_port_ha.return_value = None
                    routers = self.l3_plugin.get_sync_data_ext(adm_ctx,
                                                               [r['id']])
                    for router in routers:
                        self.assertEqual(cisco_const.ROUTER_INFO_INCOMPLETE,
                            router['status'])

    def test_populate_port_ha_information_red_router_no_port(self):

        with self.router(arg_list=(ha.ENABLED,)) as router:
            r = router['router']
            with self.port() as port:
                p = port['port']
                body = self._router_interface_action('add', r['id'], None,
                                                     p['id'])
                self.assertIn('port_id', body)
                self.assertEqual(body['port_id'], p['id'])
                adm_ctx = bc.context.get_admin_context()
                with mock.patch(
                    'networking_cisco.plugins.cisco.db.l3.ha_db.HA_db_mixin.'
                    '_populate_port_ha_information') as mock_port_ha:
                    mock_port_ha.return_value = None
                    rr_ids = [rr['id'] for rr in r[ha.DETAILS][
                        ha.REDUNDANCY_ROUTERS]]
                    routers = self.l3_plugin.get_sync_data_ext(adm_ctx, rr_ids)
                    for router in routers:
                        self.assertEqual(cisco_const.ROUTER_INFO_INCOMPLETE,
                            router['status'])

    def test_populate_port_ha_information_no_port_gw_port(self):

        with self.router(arg_list=(ha.ENABLED,)) as router:
            r = router['router']
            with self.subnet() as s:
                self._set_net_external(s['subnet']['network_id'])
                self._add_external_gateway_to_router(
                    r['id'],
                    s['subnet']['network_id'])
                adm_ctx = bc.context.get_admin_context()
                with mock.patch(
                    'networking_cisco.plugins.cisco.db.l3.ha_db.HA_db_mixin.'
                    '_populate_port_ha_information') as mock_port_ha:
                    mock_port_ha.return_value = None
                    routers = self.l3_plugin.get_sync_data_ext(adm_ctx,
                                                               [r['id']])
                    for router in routers:
                        body = self._show('routers', r['id'])
                        r_after = body['router']
                        gw_info = r_after['external_gateway_info']
                        self.assertIsNotNone(gw_info)
                        self.assertEqual(cisco_const.ROUTER_INFO_INCOMPLETE,
                            router['status'])

    def test_populate_port_ha_information_red_router_no_port_gw_port(self):

        with self.router(arg_list=(ha.ENABLED,)) as router:
            r = router['router']
            with self.subnet() as s:
                self._set_net_external(s['subnet']['network_id'])
                self._add_external_gateway_to_router(
                    r['id'],
                    s['subnet']['network_id'])
                adm_ctx = bc.context.get_admin_context()
                with mock.patch(
                    'networking_cisco.plugins.cisco.db.l3.ha_db.HA_db_mixin.'
                    '_populate_port_ha_information') as mock_port_ha:
                    mock_port_ha.return_value = None
                    rr_ids = [rr['id'] for rr in r[ha.DETAILS][
                        ha.REDUNDANCY_ROUTERS]]
                    routers = self.l3_plugin.get_sync_data_ext(adm_ctx, rr_ids)
                    for router in routers:
                        body = self._show('routers', router['id'])
                        r_after = body['router']
                        gw_info = r_after['external_gateway_info']
                        self.assertIsNotNone(gw_info)
                        self.assertEqual(cisco_const.ROUTER_INFO_INCOMPLETE,
                            router['status'])

    def test_populate_port_ha_information_retries_succeed(self):

        def fake_one():
            if m.call_count == 3:
                return hag
            elif m.call_count == 4:
                return extra_port
            else:
                raise exc.NoResultFound

        with self.router(arg_list=(ha.ENABLED,)) as router:
            r = router['router']
            with self.port() as port:
                p = port['port']
                body = self._router_interface_action('add', r['id'], None,
                                                     p['id'])
                self.assertIn('port_id', body)
                self.assertEqual(body['port_id'], p['id'])
                adm_ctx = bc.context.get_admin_context()
                hags = {}
                mod_itfcs = []
                hag = self.l3_plugin._get_ha_group_for_subnet_id(
                    adm_ctx, r['id'], p['fixed_ips'][0]['subnet_id'])
                extra_port = self._show('ports', hag.extra_port_id)['port']
                with mock.patch('sqlalchemy.orm.query.Query.one') as m:
                    m.side_effect = fake_one
                    pop_p = self.l3_plugin._populate_port_ha_information(
                        adm_ctx, p, r['id'], hags, r['id'], mod_itfcs)
                    self.assertEqual(m.call_count, 4)
                    self.assertIn(ha.HA_INFO, pop_p)
                    self.assertIsNotNone(pop_p[ha.HA_INFO]['group'])

    def test_populate_port_ha_information_all_retries_fail(self):
        with self.router(arg_list=(ha.ENABLED,)) as router:
            r = router['router']
            with self.port() as port:
                p = port['port']
                body = self._router_interface_action('add', r['id'], None,
                                                     p['id'])
                self.assertIn('port_id', body)
                self.assertEqual(body['port_id'], p['id'])
                adm_ctx = bc.context.get_admin_context()
                hags = {}
                mod_itfcs = []
                with mock.patch('sqlalchemy.orm.query.Query.one') as m:
                    m.side_effect = exc.NoResultFound
                    pop_p = self.l3_plugin._populate_port_ha_information(
                        adm_ctx, p, r['id'], hags, r['id'], mod_itfcs)
                    self.assertIsNone(pop_p)
                    self.assertEqual(len(mod_itfcs), 0)
                    self.assertEqual(m.call_count, ha_db.LOOKUP_RETRIES)

    def test_failed_add_gw_hosting_port_info_changes_router_status(self,
                                                                   num=1):
        (super(L3CfgAgentHARouterApplianceTestCase, self).
         test_failed_add_gw_hosting_port_info_changes_router_status(3))

    def test_failed_add_hosting_port_info_changes_router_status(self, num=1):
        (super(L3CfgAgentHARouterApplianceTestCase, self).
         test_failed_add_hosting_port_info_changes_router_status(3))



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""This file contains the mixin class implementation of OVS extensions for VDP.
VDP is a part of LLDP Agent Daemon (lldpad). For more information on VDP,
pls visit http://www.ieee802.org/1/pages/802.1bg.html
"""
import six

from networking_cisco.apps.saf.agent.vdp import lldpad
from networking_cisco.apps.saf.agent.vdp import vdp_constants as constants
from networking_cisco.apps.saf.common import constants as cconstants
from networking_cisco.apps.saf.common import dfa_exceptions as dfae
from networking_cisco.apps.saf.common import dfa_logger as logging
from networking_cisco.apps.saf.common import dfa_sys_lib as ovs_lib
from networking_cisco.apps.saf.common import utils as sys_utils
from neutron.agent.linux import ip_lib

LOG = logging.getLogger(__name__)


def is_uplink_already_added(root_helper, br_ex, port_name):
    br, port_exist = ovs_lib.port_exists_glob(root_helper, port_name)
    if not port_exist:
        return False
    else:
        if br == br_ex:
            return True
        else:
            LOG.error("Port %(port)s added to wrong bridge %(br)s "
                      "Given %(br_ex)s", {'port': port_name, 'br': br,
                                          'br_ex': br_ex})
            return False


def delete_uplink_and_flows(root_helper, br_ex, port_name):
    glob_delete_vdp_flows(br_ex, root_helper)
    port_exist = is_uplink_already_added(root_helper, br_ex, port_name)
    if port_exist:
        ovs_lib.delete_port_glob(root_helper, br_ex, port_name)
        lldp_ovs_veth_str = constants.LLDPAD_OVS_VETH_PORT + port_name
        ovs_lib.delete_port_glob(root_helper, br_ex, lldp_ovs_veth_str)


def glob_delete_vdp_flows(br_ex, root_helper):
    br = ovs_lib.OVSBridge(br_ex, root_helper=root_helper)
    br.delete_flows(dl_dst=constants.NCB_DMAC, dl_type=constants.LLDP_ETYPE)
    br.delete_flows(dl_dst=constants.NCB_DMAC,
                    dl_type=constants.VDP22_ETYPE)


def is_bridge_present(br, root_helper):
    ovs_bridges = ovs_lib.get_bridges(root_helper)
    if br in ovs_bridges:
        return True
    else:
        return False


class LocalVlan(object):

    def __init__(self, vlan, segmentation_id):
        self.vlan = vlan
        self.segmentation_id = segmentation_id
        self.late_binding_vlan = 0
        self.lvid = cconstants.INVALID_VLAN
        self.port_uuid_list = {}
        # Generally for the same network, there will be only one VDP VLAN
        # However, Inconsistencies can arise. This dictionary is to keep track
        # of all VLAN's floating around for different vNIC's of the same
        # network.
        self.port_vdp_vlan_dict = {}

    def set_port_uuid(self, port_uuid, vdp_vlan, fail_reason):
        if port_uuid not in self.port_uuid_list:
            port_vlan_set = [port_uuid, vdp_vlan, fail_reason]
            self.port_uuid_list[port_uuid] = port_vlan_set
            self.set_port_vlan(vdp_vlan)

    def set_portid_fail_reason(self, port_id, fail_reason):
        if port_id not in self.port_uuid_list:
            LOG.error("Unable to set fail_reason, port_uuid %s not "
                      "created", port_id)
            return
        self.port_uuid_list[port_id][2] = fail_reason

    def get_portid_fail_reason(self, port_id):
        if port_id not in self.port_uuid_list:
            return None
        return self.port_uuid_list[port_id][2]

    def get_portid_vlan(self, port_id):
        if port_id not in self.port_uuid_list:
            return cconstants.INVALID_VLAN
        return self.port_uuid_list[port_id][1]

    def set_portid_vlan(self, port_id, new_vlan):
        self.port_uuid_list[port_id][1] = new_vlan

    def set_port_vlan(self, vdp_vlan):
        if not ovs_lib.is_valid_vlan_tag(vdp_vlan):
            LOG.info("Passed Invalid vlan in set_port_vlan")
            return
        if vdp_vlan not in self.port_vdp_vlan_dict:
            self.port_vdp_vlan_dict[vdp_vlan] = 0
        self.port_vdp_vlan_dict[vdp_vlan] += 1

    def reset_port_vlan(self, vdp_vlan):
        if not ovs_lib.is_valid_vlan_tag(vdp_vlan):
            LOG.info("Passed Invalid vlan in reset_port_vlan")
            return
        if vdp_vlan not in self.port_vdp_vlan_dict:
            LOG.error("wrongly called, no VLAN's present")
            self.port_vdp_vlan_dict[vdp_vlan] = 0
        else:
            self.port_vdp_vlan_dict[vdp_vlan] -= 1
            if not self.port_vdp_vlan_dict[vdp_vlan]:
                del self.port_vdp_vlan_dict[vdp_vlan]

    def decr_reset_vlan(self, port_uuid, new_vlan):
        vlan = self.get_portid_vlan(port_uuid)
        self.reset_port_vlan(vlan)
        self.set_portid_vlan(port_uuid, new_vlan)
        self.set_port_vlan(new_vlan)

    def set_fail_reason(self, port_uuid, fail_reason):
        self.set_portid_fail_reason(port_uuid, fail_reason)

    def any_valid_vlan(self):
        return len(self.port_vdp_vlan_dict) != 0

    def any_consistent_vlan(self):
        if len(self.port_vdp_vlan_dict) != 1:
            return False
        for vlan in six.iterkeys(self.port_vdp_vlan_dict):
            return vlan
        return self.port_vdp_vlan_dict.keys()[0]


class OVSNeutronVdp(object):

    """Implements the VDP specific changes in OVS.

    Creating the veth pairs, programming the flows for VDP, deleting the VDP
    specific flows, communicating with VDP (lldpad) daemon using lldpad class
    are some of the functionality provided by this class.
    """

    def __init__(self, uplink, integ_br, ext_br, root_helper,
                 vdp_vlan_cb, vdp_mode=constants.VDP_SEGMENT_MODE):
        # self.root_helper = 'sudo'
        self.root_helper = root_helper
        self.uplink = uplink
        self.integ_br = integ_br
        self.ext_br = ext_br
        self.vdp_mode = vdp_mode
        self.local_vlan_map = {}
        self.lldpad_info = {}
        self.lldp_local_veth_port = None
        self.lldp_ovs_veth_port = None
        self.ovs_vdp_lock = sys_utils.lock()
        self.phy_peer_port_num = cconstants.INVALID_OFPORT
        self.int_peer_port_num = cconstants.INVALID_OFPORT
        self.int_peer_port = None
        self.phy_peer_port = None
        self.ext_br_obj = None
        self.integ_br_obj = None
        self.vdp_vlan_cb = vdp_vlan_cb
        self.uplink_fail_reason = ""
        self.setup_lldpad = self.setup_lldpad_ports()
        if not self.setup_lldpad:
            return
        flow_check_periodic_task = sys_utils.PeriodicTask(
            cconstants.FLOW_CHECK_INTERVAL, self._flow_check_handler)
        self.flow_check_periodic_task = flow_check_periodic_task
        flow_check_periodic_task.run()

    def is_lldpad_setup_done(self):
        return self.setup_lldpad

    def _check_bridge_flow(self, flow, out_vlan, in_vlan):
        out_vlan_flow_str = 'dl_vlan=' + str(out_vlan)
        in_vlan_flow_str = 'actions=mod_vlan_vid:' + str(in_vlan)
        flow_str = out_vlan_flow_str + ' ' + in_vlan_flow_str
        flow_partition = flow.partition(flow_str)[1]
        return len(flow_partition) != 0
        if not len(flow_partition):
            return False
        return True

    def _flow_check_handler_internal(self):
        """Periodic handler to check if installed flows are present.

        This handler runs periodically to check if installed flows are present.
        This function cannot detect and delete the stale flows, if present.
        It requires more complexity to delete stale flows. Generally, stale
        flows are not present. So, that logic is not put here.
        """
        integ_flow = self.integ_br_obj.dump_flows_for(
            in_port=self.int_peer_port_num)
        ext_flow = self.ext_br_obj.dump_flows_for(
            in_port=self.phy_peer_port_num)
        for net_uuid, lvm in six.iteritems(self.local_vlan_map):
            vdp_vlan = lvm.any_consistent_vlan()
            flow_required = False
            if not (vdp_vlan and ovs_lib.is_valid_vlan_tag(vdp_vlan)):
                return
            if not self._check_bridge_flow(integ_flow, vdp_vlan, lvm.lvid):
                LOG.error("Flow for VDP Vlan %(vdp_vlan)s, Local vlan "
                          "%(lvid)s not present on Integ bridge",
                          {'vdp_vlan': vdp_vlan, 'lvid': lvm.lvid})
                flow_required = True
            if not self._check_bridge_flow(ext_flow, lvm.lvid, vdp_vlan):
                LOG.error("Flow for VDP Vlan %(vdp_vlan)s, Local vlan "
                          "%(lvid)s not present on External bridge",
                          {'vdp_vlan': vdp_vlan, 'lvid': lvm.lvid})
                flow_required = True
            if flow_required:
                LOG.info("Programming flows for lvid %(lvid)s vdp vlan"
                         " %(vdp)s",
                         {'lvid': lvm.lvid, 'vdp': vdp_vlan})
                self.program_vm_ovs_flows(lvm.lvid, 0, vdp_vlan)

    def _flow_check_handler(self):
        """Top level routine to check OVS flow consistency. """
        LOG.info("In _flow_check_handler")
        try:
            with self.ovs_vdp_lock:
                self._flow_check_handler_internal()
        except Exception as e:
            LOG.error("Exception in _flow_check_handler_internal %s",
                      str(e))

    def program_vdp_flows(self, lldp_ovs_portnum, phy_port_num):
        br = self.ext_br_obj
        high_prio = constants.VDP_FLOW_PRIO
        br.add_flow(priority=high_prio, in_port=lldp_ovs_portnum,
                    dl_dst=constants.NCB_DMAC, dl_type=constants.LLDP_ETYPE,
                    actions="output:%s" % phy_port_num)
        br.add_flow(priority=high_prio, in_port=phy_port_num,
                    dl_dst=constants.NCB_DMAC, dl_type=constants.LLDP_ETYPE,
                    actions="output:%s" % lldp_ovs_portnum)
        br.add_flow(priority=high_prio, in_port=lldp_ovs_portnum,
                    dl_dst=constants.NCB_DMAC, dl_type=constants.VDP22_ETYPE,
                    actions="output:%s" % phy_port_num)
        br.add_flow(priority=high_prio, in_port=phy_port_num,
                    dl_dst=constants.NCB_DMAC, dl_type=constants.VDP22_ETYPE,
                    actions="output:%s" % lldp_ovs_portnum)

    def delete_vdp_flows(self):
        br = self.ext_br_obj
        br.delete_flows(dl_dst=constants.NCB_DMAC,
                        dl_type=constants.LLDP_ETYPE)
        br.delete_flows(dl_dst=constants.NCB_DMAC,
                        dl_type=constants.VDP22_ETYPE)

    def clear_obj_params(self):
        LOG.debug("Clearing Uplink Params")
        self.flow_check_periodic_task.stop()
        # How is the IP link/veth going to be removed?? fixme(padkrish)
        # IF the veth is removed, no need to unconfigure lldp/evb
        self.delete_vdp_flows()
        lldp_ovs_veth_str = constants.LLDPAD_OVS_VETH_PORT + self.uplink
        br = self.ext_br_obj
        br.delete_port(lldp_ovs_veth_str)
        br.delete_port(self.uplink)
        self.lldpad_info.clear_uplink()
        del self.lldpad_info
        # It's ok if the veth remains even if the uplink changes, worst case
        # the number of veth's will be the number of physical server ports.
        # It's not a common occurrence for uplink to change, even if so
        # the unused veth can be removed manually.
        # Reason for not removing it is the same as given in function below.
        # ip_lib.IPDevice(lldp_ovs_veth_str, self.root_helper).link.delete()

    def program_vm_ovs_flows(self, lvid, old_vlan, new_vlan):
        if old_vlan:
            # outbound
            self.ext_br_obj.delete_flows(in_port=self.phy_peer_port_num,
                                         dl_vlan=lvid)
            # inbound
            self.integ_br_obj.delete_flows(in_port=self.int_peer_port_num,
                                           dl_vlan=old_vlan)
        if new_vlan:
            # outbound
            self.ext_br_obj.add_flow(priority=4,
                                     in_port=self.phy_peer_port_num,
                                     dl_vlan=lvid,
                                     actions="mod_vlan_vid:%s,normal" %
                                             new_vlan)
            # inbound
            self.integ_br_obj.add_flow(priority=3,
                                       in_port=self.int_peer_port_num,
                                       dl_vlan=new_vlan,
                                       actions="mod_vlan_vid:%s,normal" % lvid)

    def gen_veth_str(self, const_str, intf_str):
        """Generate a veth string.

        Concatenates the constant string with remaining available length
        of interface string from trailing position.
        """
        avl_len = constants.MAX_VETH_NAME - len(const_str)
        if avl_len <= 0:
            LOG.error("veth string name too short")
            raise dfae.DfaAgentFailed(reason="Veth Unavailable")
        start_pos = len(intf_str) - avl_len
        veth_str = const_str + intf_str[start_pos:]
        return veth_str

    def setup_lldpad_ports(self):
        """Setup the flows for passing LLDP/VDP frames in OVS. """
        # Creating the physical bridge and setting up patch ports is done by
        # OpenStack
        ovs_bridges = ovs_lib.get_bridges(self.root_helper)
        if self.ext_br not in ovs_bridges or self.integ_br not in ovs_bridges:
            self.uplink_fail_reason = cconstants.bridge_not_cfgd_reason % (
                ovs_bridges, self.integ_br, self.ext_br)
            LOG.error("%s", self.uplink_fail_reason)
            raise dfae.DfaAgentFailed(reason=self.uplink_fail_reason)
        br = ovs_lib.OVSBridge(self.ext_br, root_helper=self.root_helper)
        self.ext_br_obj = br
        int_br = ovs_lib.OVSBridge(self.integ_br, root_helper=self.root_helper)
        self.integ_br_obj = int_br

        self.phy_peer_port, self.int_peer_port = self.find_interconnect_ports()
        if self.phy_peer_port is None or self.int_peer_port is None:
            self.uplink_fail_reason = cconstants.veth_not_cfgd_reason % (
                self.phy_peer_port, self.int_peer_port)
            LOG.error("%s", self.uplink_fail_reason)
            raise dfae.DfaAgentFailed(reason=self.uplink_fail_reason)
        lldp_ovs_veth_str = constants.LLDPAD_OVS_VETH_PORT + self.uplink
        if len(lldp_ovs_veth_str) > constants.MAX_VETH_NAME:
            lldp_ovs_veth_str = self.gen_veth_str(
                constants.LLDPAD_OVS_VETH_PORT,
                self.uplink)
        lldp_loc_veth_str = constants.LLDPAD_LOC_VETH_PORT + self.uplink
        if len(lldp_loc_veth_str) > constants.MAX_VETH_NAME:
            lldp_loc_veth_str = self.gen_veth_str(
                constants.LLDPAD_LOC_VETH_PORT,
                self.uplink)
        ip_wrapper = ip_lib.IPWrapper()
        self.delete_vdp_flows()
        br.delete_port(lldp_ovs_veth_str)
        if ip_lib.device_exists(lldp_ovs_veth_str):
            # What about OVS restart cases fixme(padkrish)

            # IMPORTANT.. The link delete should be done only for non-restart
            # cases. Otherwise, The MAC address of the veth interface changes
            # for every delete/create. So, if lldpad has the association sent
            # already, retriggering it will make the ASSOC appear as coming
            # from another station and more than one VSI instance will appear
            # at the Leaf. Deleting the assoc and creating the assoc for new
            # veth is not optimal. fixme(padkrish)
            # ip_lib.IPDevice(lldp_ovs_veth_str,self.root_helper).link.delete()
            lldp_loc_veth = ip_wrapper.device(lldp_loc_veth_str)
            lldp_ovs_veth = ip_wrapper.device(lldp_ovs_veth_str)
        else:
            # fixme(padkrish) Due to above reason, do the vethcreate below only
            # if it doesn't exist and not deleted.
            lldp_loc_veth, lldp_ovs_veth = (
                ip_wrapper.add_veth(lldp_loc_veth_str,
                                    lldp_ovs_veth_str))
        if not br.port_exists(self.uplink):
            phy_port_num = br.add_port(self.uplink)
        else:
            phy_port_num = br.get_port_ofport(self.uplink)
        if phy_port_num == cconstants.INVALID_OFPORT:
            self.uplink_fail_reason = cconstants.invalid_uplink_ofport_reason
            LOG.error("%s", self.uplink_fail_reason)
            return False
        if not br.port_exists(lldp_ovs_veth_str):
            lldp_ovs_portnum = br.add_port(lldp_ovs_veth)
        else:
            lldp_ovs_portnum = br.get_port_ofport(lldp_ovs_veth)
        if lldp_ovs_portnum == cconstants.INVALID_OFPORT:
            self.uplink_fail_reason = cconstants.lldp_ofport_not_detect_reason
            LOG.error("%s", self.uplink_fail_reason)
            return False
        lldp_loc_veth.link.set_up()
        lldp_ovs_veth.link.set_up()
        # What about OVS restart cases fixme(padkrish)
        self.program_vdp_flows(lldp_ovs_portnum, phy_port_num)

        self.phy_peer_port_num = br.get_port_ofport(self.phy_peer_port)
        self.int_peer_port_num = int_br.get_port_ofport(self.int_peer_port)
        if (self.phy_peer_port_num == cconstants.INVALID_OFPORT or
                self.int_peer_port_num == cconstants.INVALID_OFPORT):
            self.uplink_fail_reason = cconstants.invalid_peer_ofport_reason % (
                self.phy_peer_port_num, self.int_peer_port_num)
            LOG.error("%s", self.uplink_fail_reason)
            return False
        self.lldpad_info = (lldpad.LldpadDriver(lldp_loc_veth_str, self.uplink,
                                                self.root_helper))
        ret = self.lldpad_info.enable_evb()
        if not ret:
            self.uplink_fail_reason = cconstants.evb_cfg_fail_reason
            LOG.error("%s", self.uplink_fail_reason)
            return False
        self.lldp_local_veth_port = lldp_loc_veth_str
        self.lldp_ovs_veth_port = lldp_ovs_veth_str
        LOG.info("Setting up lldpad ports complete")
        return True

    def get_uplink_fail_reason(self):
        return self.uplink_fail_reason

    def get_lldp_local_bridge_port(self):
        return self.lldp_local_veth_port

    def get_lldp_ovs_bridge_port(self):
        return self.lldp_ovs_veth_port

    def find_interconnect_ports(self):
        """Find the internal veth or patch ports. """

        phy_port_list = self.ext_br_obj.get_port_name_list()
        int_port_list = self.integ_br_obj.get_port_name_list()
        for port in phy_port_list:
            # Use get Interface  xxx type
            is_patch = ovs_lib.is_patch(self.root_helper, port)
            if is_patch:
                # Get the peer for this patch
                peer_port = ovs_lib.get_peer(self.root_helper, port)
                if peer_port in int_port_list:
                    return port, peer_port
        # A solution is needed for veth pairs also, fixme(padkrish)
        # ip_wrapper.get_devices() returns all the devices
        # Pick the ones whose type is veth (?) and get the other pair
        # Combination of "ethtool -S xxx" command and "ip tool" command.
        return None, None

    def port_down_segment_mode(self, lldpad_port, port_uuid, mac,
                               net_uuid, segmentation_id, oui):
        lvm = self.local_vlan_map.get(net_uuid)
        if not lvm:
            fail_reason = "Local VLAN Map not available in port_down"
            LOG.error("%s", fail_reason)
            return {'result': False, 'fail_reason': fail_reason}
        if port_uuid not in lvm.port_uuid_list:
            fail_reason = "port_uuid %s not in cache for port_down" % (
                port_uuid)
            LOG.error("%s", fail_reason)
            return {'result': False, 'fail_reason': fail_reason}
        vdp_vlan = lvm.late_binding_vlan
        lldpad_port.send_vdp_vnic_down(port_uuid=port_uuid,
                                       vsiid=port_uuid,
                                       gid=segmentation_id,
                                       mac=mac, vlan=vdp_vlan, oui=oui)
        lvm.port_uuid_list.pop(port_uuid, None)
        lvm.reset_port_vlan(vdp_vlan)
        # Check here that if all the VM's in that network has
        # 0 as VLAN (dis-assoc sent by switch, but flow not removed), then
        # also remove the flow by calling unprovision. Do this after the
        # pop instruction above.
        # Check with the late binding vlan, if that still points to
        # old_vlan, remove the flow and make late_binding_vlan as 0
        # late_binding_vlan should reflect the VLAN that is installed
        # for the flow.
        if not lvm.port_uuid_list:
            self.unprovision_vdp_overlay_networks(net_uuid, lvm.lvid,
                                                  vdp_vlan, oui)
            self.local_vlan_map.pop(net_uuid)
            LOG.info("No valid ports, clearing flows")
        else:
            # There are ports present in the network case.
            if not lvm.any_valid_vlan():
                # This condition will be hit when there are still ports
                # remaining in the network, but none of them have a valid
                # VLAN. i.e. Dis-assoc sent by switch for all ports except
                # one, vdp_vlan_change will not remove flows, since there's
                # a valid port left with a VLAN. Now, user removes the VM
                # with valid port. Now flow has to be deleted since
                # there's no valid port with a VLAN.
                self.unprovision_vdp_overlay_networks(net_uuid, lvm.lvid,
                                                      vdp_vlan, oui)
                lvm.late_binding_vlan = 0
                LOG.info("unprovisioned Local %(lvid)s, VDP %(vdp)s VLAN "
                         "since no VM has valid VLAN",
                         {'lvid': lvm.lvid, 'vdp': vdp_vlan})
            else:
                # There are still valid VLAN's present.
                # Say, out of 3 VM's one VM got a VLAN change due to which
                # the new flow will be programmed according to new vlan.
                # The VM with new VLAN gets deleted.
                # Say, for whatever reason, the other VM's in the 'same'
                # network didn't gets changed to new VLAN.
                # VLAN change function won't be called and so it will
                # be stranded with stale flow unless the below
                # functionality of putting back the old VLAN is there.
                vlan_other = lvm.any_consistent_vlan()
                if vlan_other and ovs_lib.is_valid_vlan_tag(vlan_other) and (
                   vlan_other != lvm.late_binding_vlan):
                    self.program_vm_ovs_flows(lvm.lvid, vdp_vlan, vlan_other)
                    lvm.late_binding_vlan = vlan_other
                    self.vdp_nego_req = True
                    LOG.info("Reprogrammed old Flows Local %(lvid)s "
                             "VDP %(vdp)s Other %(other)s VLANs",
                             {'lvid': lvm.lvid, 'vdp': vdp_vlan,
                              'other': vlan_other})
        return {'result': True, 'fail_reason': None}

    def port_up_segment_mode(self, lldpad_port, port_name, port_uuid, mac,
                             net_uuid, segmentation_id, oui):
        lvm = self.local_vlan_map.get(net_uuid)
        if lvm and lvm.late_binding_vlan and (not lvm.vdp_nego_req):
            vdp_vlan = lvm.late_binding_vlan
            ovs_cb_data = {'obj': self, 'mac': mac,
                           'port_uuid': port_uuid, 'net_uuid': net_uuid}
            vlan, fail_reason = lldpad_port.send_vdp_vnic_up(
                port_uuid=port_uuid, vsiid=port_uuid, gid=segmentation_id,
                mac=mac, vlan=vdp_vlan, oui=oui,
                vsw_cb_fn=self.vdp_vlan_change, vsw_cb_data=ovs_cb_data)
            lvm.set_port_uuid(port_uuid, vdp_vlan, fail_reason)
            return {'result': True, 'fail_reason': fail_reason}
        else:
            int_br = self.integ_br_obj
            lvid = int_br.get_port_vlan_tag(port_name)
            if lvid != cconstants.INVALID_VLAN:
                provision_reply = self.provision_vdp_overlay_networks(
                    port_uuid, mac, net_uuid, segmentation_id, lvid, oui)
                vdp_vlan = provision_reply.get('vdp_vlan')
                if not lvm:
                    lvm = LocalVlan(lvid, segmentation_id)
                    self.local_vlan_map[net_uuid] = lvm
                lvm.lvid = lvid
                # This is just populating the list of ports in a network.
                # The vdp_vlan that's a part of port_list is just for debugging
                # So, it's ok to populate the port UUID list even if VDP VLAN
                # is invalid.
                lvm.set_port_uuid(port_uuid, vdp_vlan,
                                  provision_reply.get('fail_reason'))
                if vdp_vlan != cconstants.INVALID_VLAN:
                    lvm.late_binding_vlan = vdp_vlan
                    lvm.vdp_nego_req = False
                else:
                    LOG.error("Cannot provision VDP overlay")
                return {'result': provision_reply.get('result'),
                        'fail_reason': provision_reply.get('fail_reason')}
            else:
                fail_reason = "Invalid OVS VLAN for port %s" % (port_name)
                LOG.error("%s", fail_reason)
                return {'result': False,
                        'fail_reason': fail_reason}

    def send_vdp_port_event_internal(self, port_uuid, mac, net_uuid,
                                     segmentation_id, status, oui):
        """Send vNIC UP/Down event to VDP.

        :param port_uuid: a ovslib.VifPort object.
        :mac: MAC address of the VNIC
        :param net_uuid: the net_uuid this port is to be associated with.
        :param segmentation_id: the VID for 'vlan' or tunnel ID for 'tunnel'
        :param status: Type of port event. 'up' or 'down'
        :oui: OUI Parameters
        """
        lldpad_port = self.lldpad_info
        if not lldpad_port:
            fail_reason = "There is no LLDPad port available."
            LOG.error("%s", fail_reason)
            return {'result': False, 'fail_reason': fail_reason}

        if status == 'up':
            if self.vdp_mode == constants.VDP_SEGMENT_MODE:
                port_name = self.ext_br_obj.get_ofport_name(port_uuid)
                if port_name is None:
                    fail_reason = "Unknown portname for uuid %s" % (port_uuid)
                    LOG.error("%s", fail_reason)
                    return {'result': False, 'fail_reason': fail_reason}
                LOG.info("Status up: portname for uuid %(uuid)s is %(port)s",
                         {'uuid': port_uuid, 'port': port_name})
                ret = self.port_up_segment_mode(lldpad_port, port_name,
                                                port_uuid, mac, net_uuid,
                                                segmentation_id, oui)
        else:
            if self.vdp_mode == constants.VDP_SEGMENT_MODE:
                LOG.info("Status down for portname uuid %s", port_uuid)
                ret = self.port_down_segment_mode(lldpad_port, port_uuid,
                                                  mac, net_uuid,
                                                  segmentation_id, oui)
        return ret

    def send_vdp_port_event(self, port_uuid, mac, net_uuid,
                            segmentation_id, status, oui):
        """Send vNIC UP/Down event to VDP.

        :param port: a ovslib.VifPort object.
        :param net_uuid: the net_uuid this port is to be associated with.
        :param segmentation_id: the VID for 'vlan' or tunnel ID for 'tunnel'
        :param status: Type of port event. 'up' or 'down'
        """
        try:
            with self.ovs_vdp_lock:
                ret = self.send_vdp_port_event_internal(port_uuid, mac,
                                                        net_uuid,
                                                        segmentation_id,
                                                        status, oui)
                return ret
        except Exception as e:
            LOG.error("Exception in send_vdp_port_event %s" % str(e))
            return {'result': False, 'fail_reason': str(e)}

    def get_lvid_vdp_vlan(self, net_uuid, port_uuid):
        """Retrieve the Local Vlan ID and VDP Vlan. """

        lvm = self.local_vlan_map.get(net_uuid)
        if not lvm:
            LOG.error("lvm not yet created, get_lvid_vdp_lan "
                      "return error")
            return cconstants.INVALID_VLAN, cconstants.INVALID_VLAN
        vdp_vlan = lvm.get_portid_vlan(port_uuid)
        lvid = lvm.lvid
        LOG.info("Return from lvid_vdp_vlan lvid %(lvid)s vdp_vlan %(vdp)s",
                 {'lvid': lvid, 'vdp': vdp_vlan})
        return lvid, vdp_vlan

    def unprovision_vdp_overlay_networks(self, net_uuid, lvid, vdp_vlan, oui):
        """Unprovisions a overlay type network configured using VDP.

        :param net_uuid: the uuid of the network associated with this vlan.
        :lvid: Local VLAN ID
        :vdp_vlan: VDP VLAN ID
        :oui: OUI Parameters
        """
        # check validity
        if not ovs_lib.is_valid_vlan_tag(vdp_vlan):
            LOG.error("Cannot unprovision VDP Overlay network for"
                      " net-id=%(net_uuid)s - Invalid ",
                      {'net_uuid': net_uuid})
            return

        LOG.info('unprovision_vdp_overlay_networks: add_flow for '
                 'Local Vlan %(local_vlan)s VDP VLAN %(vdp_vlan)s',
                 {'local_vlan': lvid, 'vdp_vlan': vdp_vlan})
        self.program_vm_ovs_flows(lvid, vdp_vlan, 0)

    def vdp_vlan_change_internal(self, vsw_cb_data, vdp_vlan, fail_reason):
        """Callback Function from VDP when provider VLAN changes.

        This will be called only during error cases when switch
        reloads or when compute reloads.
        """
        LOG.debug("In VDP VLAN change VLAN %s", vdp_vlan)
        if not vsw_cb_data:
            LOG.error("NULL vsw_cb_data Info received")
            return
        net_uuid = vsw_cb_data.get('net_uuid')
        port_uuid = vsw_cb_data.get('port_uuid')
        lvm = self.local_vlan_map.get(net_uuid)
        if not lvm:
            LOG.error("Network %s is not in the local vlan map", net_uuid)
            return
        lldpad_port = self.lldpad_info
        if not lldpad_port:
            LOG.error("There is no LLDPad port available.")
            return
        exist_vdp_vlan = lvm.late_binding_vlan
        lvid = lvm.vlan
        LOG.debug("lvid %(lvid)s exist %(vlan)s",
                  {'lvid': lvid, 'vlan': exist_vdp_vlan})
        lvm.decr_reset_vlan(port_uuid, vdp_vlan)
        lvm.set_fail_reason(port_uuid, fail_reason)
        self.vdp_vlan_cb(port_uuid, lvid, vdp_vlan, fail_reason)
        if vdp_vlan == exist_vdp_vlan:
            LOG.debug("No change in provider VLAN %s", vdp_vlan)
            return
        # Logic is if the VLAN changed to 0, clear the flows only if none of
        # the VM's in the network has a valid VLAN.
        if not ovs_lib.is_valid_vlan_tag(vdp_vlan):
            if ovs_lib.is_valid_vlan_tag(exist_vdp_vlan) and not (
               lvm.any_valid_vlan()):
                # Clear the old flows
                LOG.debug("Clearing flows, no valid vlans")
                self.program_vm_ovs_flows(lvid, exist_vdp_vlan, 0)
                lvm.late_binding_vlan = 0
            lvm.vdp_nego_req = False
        else:
            # If any VM gets a VLAN change, we immediately modify the flow.
            # This is done to not wait for all VM's VLAN getting updated from
            # switch. Logic is if any VM gts a new VLAN, the other VM's of the
            # same network will be updated eventually.
            if vdp_vlan != exist_vdp_vlan and (
               ovs_lib.is_valid_vlan_tag(vdp_vlan)):
                # Add the new flows and remove the old flows
                LOG.warning("Non Zero VDP Vlan change %s %s" %
                            (vdp_vlan, exist_vdp_vlan))
                self.program_vm_ovs_flows(lvid, exist_vdp_vlan, vdp_vlan)
                lvm.late_binding_vlan = vdp_vlan
                lvm.vdp_nego_req = False
            else:
                LOG.error("Invalid or same VLAN Exist %(exist)s "
                          "New %(new)s VLANs",
                          {'exist': exist_vdp_vlan, 'new': vdp_vlan})

    def vdp_vlan_change(self, vsw_cb_data, vdp_vlan, fail_reason):
        """Callback Function from VDP when provider VLAN changes.

        This will be called only during error cases when switch
        reloads or when compute reloads.
        """
        LOG.debug("In VDP VLAN change VLAN %s" % vdp_vlan)
        try:
            with self.ovs_vdp_lock:
                self.vdp_vlan_change_internal(vsw_cb_data, vdp_vlan,
                                              fail_reason)
        except Exception as e:
            LOG.error("Exception in vdp_vlan_change %s" % str(e))

    def provision_vdp_overlay_networks(self, port_uuid, mac, net_uuid,
                                       segmentation_id, lvid, oui):
        """Provisions a overlay type network configured using VDP.

        :param port_uuid: the uuid of the VM port.
        :param mac: the MAC address of the VM.
        :param net_uuid: the uuid of the network associated with this vlan.
        :param segmentation_id: the VID for 'vlan' or tunnel ID for 'tunnel'
        :lvid: Local VLAN ID
        :oui: OUI Parameters
        """
        lldpad_port = self.lldpad_info
        if lldpad_port:
            ovs_cb_data = {'obj': self, 'port_uuid': port_uuid, 'mac': mac,
                           'net_uuid': net_uuid}
            vdp_vlan, fail_reason = lldpad_port.send_vdp_vnic_up(
                port_uuid=port_uuid, vsiid=port_uuid, gid=segmentation_id,
                mac=mac, new_network=True, oui=oui,
                vsw_cb_fn=self.vdp_vlan_change, vsw_cb_data=ovs_cb_data)
        else:
            fail_reason = "There is no LLDPad port available."
            LOG.error("%s", fail_reason)
            return {'result': False, 'vdp_vlan': cconstants.INVALID_VLAN,
                    'fail_reason': fail_reason}
        # check validity
        if not ovs_lib.is_valid_vlan_tag(vdp_vlan):
            LOG.error("Cannot provision VDP Overlay network for"
                      " net-id=%(net_uuid)s - Invalid ",
                      {'net_uuid': net_uuid})
            return {'result': True, 'vdp_vlan': cconstants.INVALID_VLAN,
                    'fail_reason': fail_reason}

        LOG.info('provision_vdp_overlay_networks: add_flow for '
                 'Local Vlan %(local_vlan)s VDP VLAN %(vdp_vlan)s',
                 {'local_vlan': lvid, 'vdp_vlan': vdp_vlan})
        self.program_vm_ovs_flows(lvid, 0, vdp_vlan)
        return {'result': True, 'vdp_vlan': vdp_vlan, 'fail_reason': None}

    def pop_local_cache(self, port_uuid, mac, net_uuid, lvid, vdp_vlan,
                        segmentation_id):
        """Populate the local cache after restart. """

        LOG.info("Populating the OVS VDP cache with port %(port_uuid)s, "
                 "mac %(mac)s net %(net_uuid)s lvid %(lvid)s vdpvlan "
                 "%(vdp_vlan)s seg %(seg)s",
                 {'port_uuid': port_uuid, 'mac': mac, 'net_uuid': net_uuid,
                  'lvid': lvid, 'vdp_vlan': vdp_vlan, 'seg': segmentation_id})
        lvm = self.local_vlan_map.get(net_uuid)
        if not lvm:
            lvm = LocalVlan(lvid, segmentation_id)
            self.local_vlan_map[net_uuid] = lvm
        lvm.lvid = lvid
        lvm.set_port_uuid(port_uuid, vdp_vlan, None)
        if vdp_vlan != cconstants.INVALID_VLAN:
            lvm.late_binding_vlan = vdp_vlan
            lvm.vdp_nego_req = False



# Copyright (c) 2017 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Stub module containing the networking_cisco trunk APIs.
#
# TODO(rcurran): Remove once networking_cisco is no longer supporting
# stable/newton.

TRUNK_SUBPORT_OWNER = ""
VLAN = ""
ACTIVE_STATUS = ""


class SubPort(object):
    @classmethod
    def get_object(cls, context, *args, **kwargs):
        return None


class TrunkObject(object):
    @classmethod
    def update(cls, **kargs):
        pass


class Trunk(object):
    @classmethod
    def get_object(cls, context, **kargs):
        return TrunkObject


class DriverBase(object):
    def __init__(self, name, interfaces, segmentation_types,
                 agent_type=None, can_trunk_bound_port=False):
        pass



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_cisco._i18n import _

from neutron_lib import exceptions as nexception

from neutron_lib.api import converters as conv

from networking_cisco import backwards_compatibility as bc

ATTR_NOT_SPECIFIED = bc.constants.ATTR_NOT_SPECIFIED
HA_ALIAS = 'router_ha'
HA = 'cisco_ha'
ENABLED = HA + ':enabled'
DETAILS = HA + ':details'
HA_INFO = 'ha_info'
TYPE = 'type'
PRIORITY = 'priority'
STATE = 'state'
HA_ACTIVE = 'ACTIVE'
HA_STANDBY = 'STANDBY'
HA_STATES = [HA_ACTIVE, HA_STANDBY]
REDUNDANCY_LEVEL = 'redundancy_level'
REDUNDANCY_ROUTERS = 'redundancy_routers'
ROUTER_ID = 'id'
PROBE_CONNECTIVITY = 'probe_connectivity'
PROBE_TARGET = 'probe_target'
PROBE_INTERVAL = 'probe_interval'
HA_VRRP = 'VRRP'
HA_HSRP = 'HSRP'
HA_GLBP = 'GLBP'
HA_TYPES = [HA_VRRP, HA_HSRP, HA_GLBP]
# number of additional routers for HA redundancy
MIN_REDUNDANCY_LEVEL = 1
MAX_REDUNDANCY_LEVEL = 3

EXTENDED_ATTRIBUTES_2_0 = {
    'routers': {
        ENABLED: {'allow_post': True, 'allow_put': True,
                  'convert_to': conv.convert_to_boolean,
                  'default': ATTR_NOT_SPECIFIED,
                  'is_visible': True},
        DETAILS: {
            'allow_post': True, 'allow_put': True,
            'is_visible': True,
            'default': ATTR_NOT_SPECIFIED,
            'enforce_policy': True,
            'validate': {
                'type:dict_or_nodata': {
                    TYPE: {'allow_post': True, 'allow_put': True,
                           'type:values': HA_TYPES,
                           'default': ATTR_NOT_SPECIFIED,
                           'is_visible': True},
                    STATE: {
                        'allow_post': False, 'allow_put': False,
                        'type:values': HA_STATES,
                        'default': ATTR_NOT_SPECIFIED,
                        'is_visible': True},
                    PRIORITY: {'allow_post': True, 'allow_put': True,
                               'convert_to': conv.convert_to_int,
                               'type:non_negative': None,
                               'default': ATTR_NOT_SPECIFIED,
                               'is_visible': True},
                    REDUNDANCY_LEVEL: {'allow_post': True, 'allow_put': True,
                                       'convert_to': conv.convert_to_int,
                                       'type:range': [MIN_REDUNDANCY_LEVEL,
                                                      MAX_REDUNDANCY_LEVEL],
                                       'default': ATTR_NOT_SPECIFIED,
                                       'is_visible': True},
                    PROBE_CONNECTIVITY: {'allow_post': True,
                                         'allow_put': True,
                                         'convert_to': conv.convert_to_boolean,
                                         'default': ATTR_NOT_SPECIFIED,
                                         'is_visible': True},
                    PROBE_TARGET: {'allow_post': True, 'allow_put': True,
                                   'type:ip_address': None,
                                   'default': ATTR_NOT_SPECIFIED,
                                   'is_visible': True},
                    PROBE_INTERVAL: {'allow_post': True, 'allow_put': True,
                                     'convert_to': conv.convert_to_int,
                                     'type:non_negative': None,
                                     'default': ATTR_NOT_SPECIFIED,
                                     'is_visible': True},
                    REDUNDANCY_ROUTERS: {
                        'allow_post': False, 'allow_put': False,
                        'is_visible': True,
                        'default': ATTR_NOT_SPECIFIED,
                        'enforce_policy': True,
                        'validate': {'type:dict_or_nodata': {
                            ROUTER_ID: {
                                'allow_post': False, 'allow_put': False,
                                'type:uuid': None},
                            STATE: {
                                'allow_post': False, 'allow_put': False,
                                'type:values': HA_STATES,
                                'default': ATTR_NOT_SPECIFIED},
                            PRIORITY: {
                                'allow_post': False, 'allow_put': False,
                                'convert_to': conv.convert_to_int,
                                'type:non_negative': None,
                                'default': ATTR_NOT_SPECIFIED}
                        }}
                    }
                }
            }
        }
    }
}


class Ha(bc.extensions.ExtensionDescriptor):
    """Extension class to support HA by VRRP, HSRP and GLBP.

    This class is used by Neutron's extension framework to support
    HA redundancy by VRRP, HSRP and GLBP for Neutron Routers.

    Attribute 'ha_type' can be one of 'vrrp', 'hsrp' and 'glbp'
    Attribute 'redundancy_level' specifies the number of routers added for
    redundancy and can be 1, 2, or 3.

    To create a router with HSRP-based HA with 2 extra routers
    for redundancy using the CLI with admin rights::

       (shell) router-create <router_name> --ha:ha_type hsrp \
       --ha:redundancy_level 2
    """

    @classmethod
    def get_name(cls):
        return "High-availability for routing service"

    @classmethod
    def get_alias(cls):
        return HA_ALIAS

    @classmethod
    def get_description(cls):
        return "High availability by VRRP, HSRP, and GLBP"

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/ha/api/v1.0"

    @classmethod
    def get_updated(cls):
        return "2013-12-07T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return EXTENDED_ATTRIBUTES_2_0
        else:
            return {}


# HA exceptions
class HADisabled(nexception.Conflict):
    message = _("HA support is disabled")


class HADisabledHAType(nexception.Conflict):
    message = _("HA type %(ha_type)s is administratively disabled")


class HARedundancyLevel(nexception.BadRequest):
    message = _("Redundancy level for HA must be 1, 2, or 3")


class HATypeCannotBeChanged(nexception.Conflict):
    message = _("HA type cannot be changed for a router with HA enabled")


class HATypeNotCompatibleWithFloatingIP(nexception.BadRequest):
    message = _("HA type %(ha_type)s cannot be used with FloatingIP")



# Copyright 2015 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


import datetime
import os
import six
import socket
import struct
import sys
import threading
from threading import Lock
import time
import traceback
import uuid


TIME_FORMAT = '%a %b %d %H:%M:%S %Y'


class PeriodicTask(object):

    """Periodic task"""

    def __init__(self, interval, func, **kwargs):
        self._interval = interval
        self._fn = func
        self._kwargs = kwargs
        self.stop_flag = False
        self._excq = kwargs.get('excq')

    def run(self):
        try:
            if self.stop_flag:
                return
            start = time.time()
            self._fn(**self._kwargs)
            end = time.time()
            delta = end - start
            self.thrd = threading.Timer(self._interval - delta, self.run)
            self.thrd.start()
        except Exception as e:
            if self._excq:
                emsg = ('%(name)s : %(excp)s' % {'name': self._fn.__name__,
                                                 'excp': str(e)})
                self._excq.put(emsg, block=False)

    def stop(self):
        try:
            self.thrd.cancel()
            self.stop_flag = True
        except Exception as e:
            if self._excq:
                emsg = ('Exception in timer stop %s' % str(e))
                self._excq.put(emsg, block=False)


class EventProcessingThread(threading.Thread):

    """Event processing thread."""

    def __init__(self, name, obj, task, excq=None):
        super(EventProcessingThread, self).__init__(name=name)
        self._thread_name = name
        self._hdlr = obj
        self._task = task
        self._excq = excq

    def run(self):
        try:
            getattr(self._hdlr, self._task)()
        except Exception:
            if self._excq:
                exc_type, exc_value, exc_tb = sys.exc_info()
                tbstr = traceback.format_exception(exc_type, exc_value, exc_tb)
                exstr = str(dict(name=self._thread_name, tb=tbstr))
                self._excq.put(exstr, block=False)

    @property
    def am_i_active(self):
        return self.isAlive()

    @property
    def name(self):
        return self._thread_name


class Dict2Obj(object):

    """Convert a dictionary to an object."""

    def __init__(self, d):
        for key, val in six.iteritems(d):
            # Check if it is nested dictionary
            if isinstance(val, dict):
                setattr(self, key, Dict2Obj(val))
            else:
                setattr(self, key, val)

    def __getattr__(self, val):
        return self.__dict__.get(val)


def get_uuid():
    return str(uuid.uuid4())


def lock():
    return Lock()


def utc_time(ct):
    if ct:
        return datetime.datetime.strptime(ct, TIME_FORMAT)


def utc_time_lapse(lapse):
    current_time = utc_time(time.ctime())
    hour_lapse = current_time - datetime.timedelta(hours=lapse)
    return hour_lapse


def is_valid_ipv4(addr):
    try:
        socket.inet_aton(addr)
        return True
    except socket.error:
        return False


def is_valid_mac(addr):
    """Check the syntax of a given mac address.

    The acceptable format is xx:xx:xx:xx:xx:xx
    """
    addrs = addr.split(':')
    if len(addrs) != 6:
        return False
    for m in addrs:
        try:
            if int(m, 16) > 255:
                return False
        except ValueError:
            return False
    return True


def make_cidr(gw, mask):
    """Create network address in CIDR format.

    Return network address for a given gateway address and netmask.
    """
    try:
        int_mask = (0xFFFFFFFF << (32 - int(mask))) & 0xFFFFFFFF
        gw_addr_int = struct.unpack('>L', socket.inet_aton(gw))[0] & int_mask
        return (socket.inet_ntoa(struct.pack("!I", gw_addr_int)) +
                '/' + str(mask))
    except (socket.error, struct.error, ValueError, TypeError):
        return


def find_agent_host_id(this_host):
    """Returns the neutron agent host id for RHEL-OSP6 HA setup."""

    host_id = this_host
    try:
        for root, dirs, files in os.walk('/run/resource-agents'):
            for fi in files:
                if 'neutron-scale-' in fi:
                    host_id = 'neutron-n-' + fi.split('-')[2]
                    break
        return host_id
    except IndexError:
        return host_id
